#ifndef _WON_H
#define _WON_H

#ifndef MAX_PATH
#define MAX_PATH	512
#endif

#include "bitwise.h"
#include "list.h"
#include "hash.h"
#include "cache.h"
#include "md5.h"

#include <sys/socket.h>
#include <netinet/in.h>

typedef unsigned long uint32;
typedef unsigned long WON_STATUS;

#define PRINT_IP(addr)          ((unsigned char *)&addr)[0],                                            \
                                ((unsigned char *)&addr)[1],                                            \
				((unsigned char *)&addr)[2],                                            \
				((unsigned char *)&addr)[3]



/* Maximums 			*/

#define WON_MAX_MODULENAME_SIZE	0x00000100
#define WON_MAX_CONNECTIONS	0x00001000
#define WON_MAX_NICKNAME_SIZE	0x00000010
#define WON_MAX_MASK_SIZE	0x00000100

/* IOCTL defines 		*/

#define WON_IOCTL_CONFIG_SAVE	0x00000001	/* Configuration file is being saved noficiation */
#define WON_IOCTL_IRC_LINK_UP	0x00000002	/* IRC Connection link is up			 */
#define WON_IOCTL_IRC_LINK_DOWN	0x00000003	/* IRC Connection link is down			 */
#define WON_IOCTL_DCC_LINK_UP	0x00000004	/* DCC Connection link is up			 */
#define WON_IOCTL_DCC_LINK_DOWN	0x00000005	/* DCC Connection link is down			 */
#define WON_IOCTL_LOAD_COMPLETE 0x00000006	/* After all modules are loaded this ioctl used  */

/* Common return codes 		*/

#define WON_STATUS_SUCCESS	0x00000000	/* The operation completed successfully		 */
#define WON_STATUS_FAILURE	0x00000001	/* A generic failure occured			 */
#define WON_STATUS_EXISTS	0x00000002	/* The object already exists			 */
#define WON_STATUS_NOEXISTS	0x00000003	/* The object does not exist			 */
#define WON_STATUS_CLAIMPACKET	0x00000004	/* Claim the packet and do not pass down	 */
#define WON_STATUS_CONNCLOSED	0x00000005	/* The connection has been closed		 */

/* Connection types		*/

#define WON_CONNECTION_BOTNET	0x00000000	/* Botnet connection				 */
#define WON_CONNECTION_DCC	0x00000001	/* DCC connection				 */
#define WON_CONNECTION_IRC	0x00000002	/* IRC connection				 */

/* Layering preferences		*/

#define WON_LAYER_FIRST		0x00000001	/* Layer first					 */
#define WON_LAYER_LAST		0x00000000	/* layer last					 */

/* States			*/

#define WON_IRC_STATE_CONNECTED	(1 << 1)	/* TCP connection established			 */
#define WON_IRC_STATE_ONLINE	(1 << 2)	/* Online with a nickname			 */

#define WON_DCC_STATE_USERNAME	(1 << 1)	/* Username has been recv'd			 */
#define WON_DCC_STATE_AUTH	(1 << 2)	/* User is authenticated			 */

#define WON_SOCKET_STATE_CONNECTED	(1 << 1)

/* Low level operations		*/

#define WON_OPERATION_SEND	0x01
#define WON_OPERATION_RECV	0x02

/* Defaults			*/

#define WON_DEFAULT_VERSION_REPLY	"wonton -- the definitive answer."
#define WON_DEFAULT_DCC_USER_PROMPT	"username"
#define WON_DEFAULT_DCC_PASS_PROMPT	"password"

/* Core wonton structure 	*/

typedef struct _won_connection {
	unsigned short 	   type;		/* Connection type, DCC, IRC, BOTNET	*/
	int 		   fd;			/* Connection fd			*/
	int		   state;		/* Socket state				*/
	struct sockaddr_in addr;		/* Addr info				*/

	CACHE		   buffer;		/* Buffer for the connection		*/
} WON_CONNECTION;

typedef struct _won_ctx {
	/* Core 	*/
	short		initialized;
	uint32	 	startTime;
	uint32	 	moduleCount;
	short		wantComplete;
	uint32		pid;

	char 		*configurationFile;	/* Configuration file location		*/

	WON_CONNECTION 	wonConnections[512];	/* Array of connections			*/
	uint32		numConnections;		/* Number of connections		*/
	int		highestFd;		/* The highest fd			*/

	LINKED_LIST 	moduleList;		/* List of loaded modules		*/
	HASH_GROUP  	moduleHash;		/* Hash of loaded modules		*/

	LINKED_LIST 	ircConnections;		/* List of current irc connections	*/
	LINKED_LIST 	dccConnections;		/* List of current dcc connections	*/

	LINKED_LIST 	schedulerList;		/* List of scheduled tasks		*/

	/* Handlers 	*/

	HASH_GROUP	ircHandlers;		/* Hash of WON_HANDLERS			*/
	HASH_GROUP	dccHandlers;		/* Hash of WON_HANDLERS			*/
	HASH_GROUP	configHandlers;		/* Hash of WON_HANDLERS			*/
	LINKED_LIST	lowlevelRecvHandlers;	/* low level recv handlers		*/
	LINKED_LIST	lowlevelSendHandlers;	/* low level send handlers		*/

	FILE		*logfd;			/* Logging file descriptor		*/
	char		logFile[MAX_PATH];	/* Logging file				*/
} WON_CTX;

typedef struct _won_module_ctx {
	WON_CTX		*wonctx;		/* The wonton context			*/
	void 		*moduleHandle;		/* The module handle			*/
	void 		*moduleExtension;	/* The module extension			*/
	
	uint32		(*moduleLoad)(struct _won_module_ctx *modctx);
	uint32		(*moduleIoctl)(struct _won_module_ctx *modctx, uint32 ioctl, void *wic);
	uint32		(*moduleUnload)(struct _won_module_ctx *modctx);

	char		moduleName[WON_MAX_MODULENAME_SIZE];
	char		modulePath[MAX_PATH];
	uint32		moduleSize;
} WON_MODULE_CTX;

typedef struct _won_interrupt_ctx {
	WON_MODULE_CTX	*modctx;		/* The module handler coming in		*/

	void		*inBuffer;		/* The in buffer of the request		*/
	uint32		inLen;			/* The in buffer length			*/

	void		*outBuffer;		/* The out buffer of the request	*/
	uint32		outLen;
} WON_IC;

typedef struct _won_scheduled_task {
	WON_CTX		*ctx;			/* Context				*/
	WON_MODULE_CTX	*modctx;		/* Module context			*/

	uint32		id;			/* Scheduled tasks ID			*/
	uint32		timeFrame;		/* Time between executes		*/
	uint32		executeTimes;		/* Number of times to execute		*/

	void 		*schedulerContext;	/* Tasks context			*/

	uint32		(*callback)(struct _won_scheduled_task *task);
	
	uint32		lastExecute;		/* last executed time			*/
} WON_SCHEDULED_TASK;

typedef struct _won_irc_connection {
	unsigned long 	addr;
	unsigned int 	port;
	int		fd;

	unsigned int	state;

	char 		nickname[WON_MAX_NICKNAME_SIZE];

	void 		*ircExtension;
} WON_IRC_CONNECTION;

typedef struct _won_dcc_connection {
	int		fd;
	unsigned long	addr;
	unsigned int 	port;
	
	unsigned int	state;

	void		*dccExtension;
} WON_DCC_CONNECTION;

typedef struct _won_irc_packet {
	WON_CTX			*ctx;		/* Core context					*/
	WON_MODULE_CTX 		*modctx;	/* The module the packet is being passed to 	*/
	WON_IRC_CONNECTION 	*irc;		/* The irc connection the packet came with	*/

	CACHE			*buffer;	/* Wonton parsed buffer				*/
	char			fromNick[WON_MAX_NICKNAME_SIZE];
	char			fromMask[WON_MAX_MASK_SIZE];
} WON_IRC_PACKET;

typedef struct _won_dcc_packet {
	WON_CTX			*ctx;		/* Core context					*/
	WON_MODULE_CTX 		*modctx;	/* The module the packet is being passed to 	*/
	WON_DCC_CONNECTION 	*dcc;		/* The dcc connection the packet came with	*/

	CACHE			*buffer;	/* Wonton parsed buffer				*/
} WON_DCC_PACKET;

typedef struct _won_config_packet {
	WON_CTX			*ctx;		/* Core context					*/
	WON_MODULE_CTX		*modctx;	/* The module the packet is being passed to	*/
	
	CACHE			*buffer;	/* Wonton parsed buffer				*/

	uint32			lineNumber;	/* Line number					*/
} WON_CONFIG_PACKET;

typedef struct _won_packet {
	short			type;

	union {
		WON_IRC_PACKET 	*irc;
		WON_DCC_PACKET	*dcc;
	} pkt;
} WON_PACKET;

typedef struct _won_registered_handler {
	WON_MODULE_CTX	*modctx;
	uint32		function;
	int		position;

	struct _won_registered_handler *prev;
	struct _won_registered_handler *next;
} WON_REGISTERED_HANDLER;

typedef struct _won_handlers {
	unsigned char 	*item;
	
	WON_REGISTERED_HANDLER *firstHandler;
	WON_REGISTERED_HANDLER *lastHandler;
} WON_HANDLERS;


/**** Core functions ****/

/* Memory Management 		*/

void *wonMemoryAllocate(uint32 size,short flags);
void *wonMemoryReallocate(void *ptr, uint32 size);
void wonMemoryFree(void *ptr);

/* Module Management 		*/

WON_STATUS wonModuleLoad(WON_CTX *ctx,
				const char *modulePath,
				const char *moduleName);
WON_STATUS wonModuleIoctl(WON_CTX *ctx,
				const char *moduleName,
				uint32 ioctl,
				WON_IC *wic);
WON_STATUS wonModuleUnload(WON_CTX *ctx,
				const char *moduleName);
WON_STATUS wonModuleDeferredUnload(WON_CTX *ctx,
				const char *moduleName);

void wonModuleCleanup(WON_CTX *ctx,
				WON_MODULE_CTX *modctx);
WON_STATUS wonModuleBroadcastIoctl(WON_CTX *ctx,
				uint32 ioctl,
				void *inBuffer,
				uint32 inLen);

/* Low level			*/

WON_STATUS wonLowlevelRecvRegisterLayeredHandler(WON_CTX *ctx,
						WON_MODULE_CTX *modctx,
						short layerPreference,
						uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet));
WON_STATUS wonLowlevelRecvDeregisterLayeredHandler(WON_CTX *ctx,
						uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet));
WON_STATUS wonLowlevelSendRegisterLayeredHandler(WON_CTX *ctx,
						WON_MODULE_CTX *modctx,
						short layerPreference,
						uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet));
WON_STATUS wonLowlevelSendDeregisterLayeredHandler(WON_CTX *ctx,
						uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet));


/* Scheduler Management 	*/

WON_STATUS wonSchedulerRegister(WON_CTX *ctx,
				WON_MODULE_CTX *modctx,
				uint32 *schedulerId,
				uint32 timeFrame,
				uint32 executeTimes,
				uint32 (*func)(WON_SCHEDULED_TASK *task),
				void *schedulerContext);
WON_STATUS wonSchedulerDeregisterById(WON_CTX *ctx,
				  uint32 schedulerId);
WON_STATUS wonSchedulerDeregisterByCallback(WON_CTX *ctx,
				  	uint32 (*func)(WON_SCHEDULED_TASK *task));

/* Networking (IRC)		*/

WON_STATUS wonIrcSend(WON_CTX *ctx,
				WON_IRC_CONNECTION *irc,
				const char *fmt,
				...);
WON_STATUS wonIrcEnum(WON_CTX *ctx,
				uint32 ircConnection,
				WON_IRC_CONNECTION **irc);
WON_STATUS wonIrcCreateConnection(WON_CTX *ctx,
				unsigned long bindaddr,
				unsigned char *dst,
				uint32 port);
WON_STATUS wonIrcDestroyConnection(WON_CTX *ctx,
				WON_IRC_CONNECTION *irc);

WON_IRC_CONNECTION *wonIrcGetConnectionByHandle(WON_CTX *ctx,
						int fd);

void wonIrcConnectionSetNickName(WON_CTX *ctx,
				WON_IRC_CONNECTION *irc,
				const char *nickname);
#define wonIrcConnectionGetNickName(irc) (irc)->nickname
#define wonIrcConnectionSetState(irc,statez) BIT_SET((irc)->state,statez)
#define wonIrcConnectionIsState(irc,statez) IS_SET((irc)->state,statez)
#define wonIrcConnectionUnsetState(irc,statez) BIT_REMOVE((irc)->state,statez)

#define wonIrcRegisterLayeredHandler(ctx,modctx,layerPreference,item,position,func) wonRegisterLayeredHandler(&(ctx)->ircHandlers,ctx,modctx,layerPreference,item,position,func)
#define wonIrcDeregisterLayeredHandler(ctx,modctx,item,position,func) wonDeregisterLayeredHandler(&(ctx)->ircHandlers,ctx,modctx,item,position,func)

/* Networking (DCC) 		*/

WON_STATUS wonDccRegisterConnection(WON_CTX *ctx,
				WON_DCC_CONNECTION *dcc);

WON_STATUS wonDccDeregisterConnection(WON_CTX *ctx,
				WON_DCC_CONNECTION *dcc);

WON_STATUS wonDccSend(WON_CTX *ctx,
				WON_DCC_CONNECTION *dcc,
				const char *fmt,
				...);
WON_STATUS wonDccEnum(WON_CTX *ctx,
				uint32 dccConnection,
				WON_DCC_CONNECTION **dcc);
WON_STATUS wonDccSendAll(WON_CTX *ctx,
				const char *fmt,
				...);

WON_DCC_CONNECTION *wonDccGetConnectionByHandle(WON_CTX *ctx,
							int fd);

#define wonDccRegisterLayeredHandler(ctx,modctx,layerPreference,item,position,func) wonRegisterLayeredHandler(&(ctx)->dccHandlers,ctx,modctx,layerPreference,item,position,func)
#define wonDccDeregisterLayeredHandler(ctx,modctx,item,position,func) wonDeregisterLayeredHandler(&(ctx)->dccHandlers,ctx,modctx,item,position,func)

/* Configuration 		*/

WON_STATUS wonConfigurationSave(WON_CTX *ctx);
WON_STATUS wonConfigurationReload(WON_CTX *ctx);

#define wonConfigRegisterLayeredHandler(ctx,modctx,layerPreference,item,func) wonRegisterLayeredHandler(&(ctx)->configHandlers,ctx,modctx,layerPreference,item,0,func)
#define wonConfigDeregisterLayeredHandler(ctx,modctx,item,func) wonDeregisterLayeredHandler(&(ctx)->configHandlers,ctx,modctx,item,0,func)

/* Logging			*/

WON_STATUS wonLog(WON_CTX *ctx,
		WON_MODULE_CTX *modctx,
		const char *fmt,
		...);

/* Helper functions */

WON_STATUS wonRegisterLayeredHandler(HASH_GROUP *hash,
					WON_CTX *ctx,
					WON_MODULE_CTX *modctx,
					short layerPreference,
					const char *item,
					int position,
					void *func);
WON_STATUS wonDeregisterLayeredHandler(HASH_GROUP *hash,
					WON_CTX *ctx,
					WON_MODULE_CTX *modctx,
					const char *item,
					int position,
					void *func);

void wonConnectionAdd(WON_CTX *ctx,
			unsigned short type,
			int fd,
			struct sockaddr_in *s);
void wonConnectionDelete(WON_CTX *ctx,
			unsigned short type,
			int fd);

int wonWildcardMatch(unsigned char *wildcardStr,
			unsigned char *realStr);

int wonCreateTcpConnection(unsigned long bindaddr,
				const char *dst,
				unsigned int port,
				unsigned long *resolveaddr);

#endif
