/*
 * libcache
 *
 * Author: Matt Miller
 *		(mmiller@hick.org)
 *
 * String parsing library.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef SUNOS
	#include <stropts.h>
	#define NREAD 	I_NREAD
#else
	#define NREAD 	FIONREAD
#endif
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>
#include "cache.h"

int cacheParseWords(CACHE *cache)
{
	char tmpword[BUF_SIZE];
        int count, pos = 0;

	if (!cache)
		return E_BAD_CACHE;

        if (!cacheFree(cache))
		return E_ERROR;

	memset(tmpword,0,BUF_SIZE);

	if (!cache->bufpos)
		return E_ERROR;

	for (count = 0;count <= cache->bufpos;count++)
	{
		if ((cache->buf[count] == '\r') || (cache->buf[count] == '\n'))
			continue;
		else if (((cache->buf[count] == ' ') && (cache->buf[count-1] != ' ')) || 
			((cache->buf[count] == ' ') && (cache->buf[count-1] != '\t')) ||
			((cache->buf[count] == '\t') && (cache->buf[count-1] != ' ')) ||
			((cache->buf[count] == '\t') && (cache->buf[count-1] != '\t')) ||
			(count == cache->bufpos))
		{
			tmpword[pos] = 0;

			if (tmpword[0] != 0)
			{
				if ((cache->words+1) == CACHE_SIZE)
					break;

				cache->str[cache->words++] = strdup(tmpword);

				memset(tmpword,0,pos);
				pos = 0;
			}
		}
		else if ((cache->buf[count] == ' ') || (cache->buf[count] == '\t'))
			continue;
		else
			tmpword[pos++] = cache->buf[count];
	}

	return S_OK;
}

int cacheFree(CACHE *cache)
{
        int count;

	if (!cache)
		return E_BAD_CACHE;

        for (count = 0;((count <= cache->words) && (count < CACHE_SIZE));count++)
        {
                if (cache->str[count] != NULL)
                {
                        free(cache->str[count]);
                        cache->str[count] = NULL;
                }
        }
        cache->words = 0;

        return S_OK;
}

int cacheInitialize(CACHE *cache)
{
        int count;

        if (!cache)
                return E_BAD_CACHE;

        for (count = 0;count < CACHE_SIZE;count++)
                cache->str[count] = NULL;

        cache->words = 0;

        memset(cache->buf,0,BUF_SIZE);
        cache->bufpos = 0;

        return S_OK;
}

int cacheReadFileFDLine(FILE *fd, CACHE *cache)
{
	if (!fd)
		return E_FD_CLOSED;

	if (!cache)
		return E_BAD_CACHE;

	fgets(cache->buf,(BUF_SIZE-1),fd);

	cache->bufpos = strlen(cache->buf);

	cacheParseWords(cache);

	return 1;
}

int cacheReadSockFD(int fd, CACHE *cache, int action)
{
	int val;

	switch (action)
	{
		case A_READLNP: case A_READLN:
		
			while ((val = _cacheReadSock(fd,cache,action)) != S_GOTEOLN)
			{
				if (!val)
					return val;
			}

			if (action == A_READLNP)
				cacheParseWords(cache);

			return val;

		case A_READP: case A_READ:
	
			val = _cacheReadSock(fd,cache,action);

			if (val)
			{
				if (action == A_READP)
					val = cacheParseWords(cache);
			}

			return val;

		default:

			return E_BAD_ACTION;
			
	}

	return E_ERROR;
}

int cacheAddCharacter(CACHE *cache, char character)
{
	if (!cache)
		return E_BAD_CACHE;

	if ((cache->bufpos+1) == BUF_SIZE)
		return E_CACHE_BUFFER_FULL;

	cache->buf[cache->bufpos++] = character;

	return S_OK;
}

int _cacheReadSock(int fd, CACHE *cache, int action)
{
	int retval, nread = 0, err;
	struct timeval tv;	
	fd_set fds;
	char tbuf[2] = { 0, 0 };

	tv.tv_sec = 0;
	tv.tv_usec = 500;

	
	FD_ZERO(&fds);
	FD_SET(fd, &fds);

	retval = select(fd + 1, &fds, NULL, NULL, &tv);

	if (retval)
	{
		nread = 1;

		while (nread)
		{
			ioctl(fd, NREAD, &nread);

			if (!nread)
				break;

			if (!read(fd,tbuf,1))
				return E_FD_CLOSED;
		
			if (!(err = cacheAddCharacter(cache,tbuf[0])))
				return err; 

			if ((action == A_READLNP) || (action == A_READLN))
				if (tbuf[0] == '\n')
					return S_GOTEOLN;
		}
	}

	return S_OK;
}

void cacheCompileWords(CACHE *cache, int start, int finish, char *delim, char *retbuf, unsigned int retbufsize)
{
	int c = start, retsize = 0, tlen = 0;

	if (!cache)
		return;

	if (start >= cache->words)
		return;

	memset(retbuf,0,retbufsize);

	retsize = retbufsize;

	for (;((c < cache->words) && (c < finish));c++)
	{
		if (!cache->str[c])
			continue;

		tlen = strlen(cache->str[c]);

		if ((retsize - tlen) <= 2)
			break;

		if (c == start)
			strcpy(retbuf,cache->str[c]);
		else
			strcat(retbuf,cache->str[c]);

		if (cache->str[c+1])
			strcat(retbuf,delim);

		retsize -= tlen;
	}

	return;
}
