/* This file consists of exported functions for modules */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <dlfcn.h>
#include <string.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>

#include "won.h"
#include "wonloader.h"

uint32 _wonInternalModuleUnload(WON_SCHEDULED_TASK *task);

/* Memory Management */

void *wonMemoryAllocate(uint32 size,short zeroMemory)
{
	void *retptr = NULL;

	while (!(retptr = malloc(size)))
		sleep(1);

	if (zeroMemory)
		memset(retptr,0,size);

	return retptr;
}

void *wonMemoryReallocate(void *ptr, uint32 size)
{
	void *retptr;

	while (!(retptr = realloc(ptr,size)))
		sleep(1);

	return retptr;
}

void wonMemoryFree(void *ptr)
{
	if (ptr)
		free(ptr);
}

/* Module Management */

WON_STATUS wonModuleLoad(WON_CTX *ctx,
				const char *modulePath,
				const char *moduleName)
{
	WON_MODULE_CTX *modctx;
	void *modHandle;
	struct stat statbuf;

	modHandle = dlopen(modulePath,RTLD_NOW);

	if (hashStringKeyGetValue(&ctx->moduleHash,moduleName))
		return WON_STATUS_EXISTS;

	if (!modHandle)
		return WON_STATUS_FAILURE;

	stat(modulePath,&statbuf);

	modctx = wonMemoryAllocate(sizeof(WON_MODULE_CTX),1);

	modctx->moduleLoad	= (uint32 (*)(WON_MODULE_CTX *))dlsym(modHandle,"wonLoad");
	modctx->moduleIoctl	= (uint32 (*)(WON_MODULE_CTX *, uint32 ioctl, void *wic))dlsym(modHandle,"wonIoctl");
	modctx->moduleUnload	= (uint32 (*)(WON_MODULE_CTX *))dlsym(modHandle,"wonUnload");

	modctx->wonctx		= ctx;
	modctx->moduleHandle 	= modHandle;
	modctx->moduleSize	= statbuf.st_size / 1024;
	strncpy(modctx->moduleName,moduleName,WON_MAX_MODULENAME_SIZE-1);
	strncpy(modctx->modulePath,modulePath,MAX_PATH-1);

	hashStringKeyAdd(&ctx->moduleHash,modctx->moduleName,modctx);
	listAddStringId(&ctx->moduleList,modctx->moduleName,modctx);

	ctx->moduleCount++;

	if (modctx->moduleLoad)
		modctx->moduleLoad(modctx);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonModuleUnload(WON_CTX *ctx,
				const char *moduleName)
{
	WON_MODULE_CTX *modctx;

	if (!(modctx = (WON_MODULE_CTX *)hashStringKeyGetValue(&ctx->moduleHash,moduleName)))
		return WON_STATUS_NOEXISTS;

	if ((modctx->moduleUnload) && (modctx->moduleUnload(modctx) == WON_STATUS_FAILURE))
		return WON_STATUS_FAILURE;

	/* Clean up everything related to this module just in case... */

	wonModuleCleanup(ctx,modctx);	

	dlclose(modctx->moduleHandle);

	hashStringKeyDelete(&ctx->moduleHash,moduleName);
	listDeleteStringId(&ctx->moduleList,moduleName);

	ctx->moduleCount--;

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonModuleDeferredUnload(WON_CTX *ctx,
                                const char *moduleName)
{
	char *newModuleName = strdup(moduleName);

	return wonSchedulerRegister(ctx,
				NULL,
				NULL,
				1,
				1,
				_wonInternalModuleUnload,
				newModuleName);
}

WON_STATUS _wonInternalModuleUnload(WON_SCHEDULED_TASK *task)
{
	wonModuleUnload(task->ctx,task->schedulerContext);

	wonMemoryFree(task->schedulerContext);

	return WON_STATUS_SUCCESS;
}

void wonModuleCleanup(WON_CTX *ctx, WON_MODULE_CTX *modctx)
{
	LINKED_LIST_ITEM *item;
	WON_SCHEDULED_TASK *task;
	HASH_ENTRY *entry;
	int x = 0;

	/* Scheduled Tasks */

	item = listGetFirstItem(&ctx->schedulerList);

	while (item && item->value)
	{
		task = (WON_SCHEDULED_TASK *)item->value;

		if (task->modctx == modctx)
		{
			wonLog(ctx,NULL,"moduleCleanup: module %s missed dregister of scheduled task.  cleaned.\n",modctx->moduleName);

			wonSchedulerDeregisterById(ctx,task->id);

			item = listGetFirstItem(&ctx->schedulerList);

			continue;
		}

		item = listGetNextItem(item);
	}

	/* Config Handlers */

	for (x = 0; x < ctx->configHandlers.hashBuckets; x++)
	{
		entry = ctx->configHandlers.hashTable[x].hashEntry;
		while (entry)
		{
			WON_HANDLERS *hndl = (WON_HANDLERS *)entry->ptr;
			WON_REGISTERED_HANDLER *curr = (hndl)?hndl->firstHandler:NULL;

			while (curr)
			{
				if (curr->modctx == modctx)
				{
					wonLog(ctx,NULL,"moduleCleanup: module %s missed deregister for config item '%s'.  cleaned.",modctx->moduleName,hndl->item);
					wonConfigDeregisterLayeredHandler(ctx,
										modctx,
										hndl->item,
										(uint32 (*)(WON_CONFIG_PACKET *))curr->function);
					break;
				}

				curr = curr->next;
			}

			if (!curr)
				entry = entry->next;
			else
				entry = ctx->configHandlers.hashTable[x].hashEntry;
		}
	}

	/* DCC Handlers */

	for (x = 0; x < ctx->dccHandlers.hashBuckets; x++)
	{
		entry = ctx->dccHandlers.hashTable[x].hashEntry;
		while (entry)
		{
			WON_HANDLERS *hndl = (WON_HANDLERS *)entry->ptr;
			WON_REGISTERED_HANDLER *curr = (hndl)?hndl->firstHandler:NULL;

			while (curr)
			{
				if (curr->modctx == modctx)
				{
					wonLog(ctx,NULL,"moduleCleanup: module %s missed deregister for dcc item '%s'.  cleaned.",modctx->moduleName,hndl->item);
					wonDccDeregisterLayeredHandler(ctx,
										modctx,
										hndl->item,
										curr->position,
										(uint32 (*)(WON_DCC_PACKET *))curr->function);
					break;
				}

				curr = curr->next;
			}

			if (!curr)
				entry = entry->next;
			else
				entry = ctx->dccHandlers.hashTable[x].hashEntry;
		}
	}

	/* IRC Handlers */

	for (x = 0; x < ctx->ircHandlers.hashBuckets; x++)
	{
		entry = ctx->ircHandlers.hashTable[x].hashEntry;
		while (entry)
		{
			WON_HANDLERS *hndl = (WON_HANDLERS *)entry->ptr;
			WON_REGISTERED_HANDLER *curr = (hndl)?hndl->firstHandler:NULL;

			while (curr)
			{
				if (curr->modctx == modctx)
				{
					wonLog(ctx,NULL,"moduleCleanup: module %s missed deregister for irc item '%s'.  cleaned.",modctx->moduleName,hndl->item);
					wonIrcDeregisterLayeredHandler(ctx,
										modctx,
										hndl->item,
										curr->position,
										(uint32 (*)(WON_DCC_PACKET *))curr->function);
					break;
				}

				curr = curr->next;
			}

			if (!curr)
				entry = entry->next;
			else
				entry = ctx->ircHandlers.hashTable[x].hashEntry;
		}
	}
}

WON_STATUS wonModuleIoctl(WON_CTX *ctx,
				const char *moduleName,
				uint32 ioctl,
				WON_IC *wic)
{
	WON_MODULE_CTX *modctx;

	if (!(modctx = (WON_MODULE_CTX *)hashStringKeyGetValue(&ctx->moduleHash,moduleName)))
		return WON_STATUS_NOEXISTS;

	return (modctx->moduleIoctl) ? modctx->moduleIoctl(modctx,ioctl,wic) : WON_STATUS_FAILURE;
}

WON_STATUS wonModuleBroadcastIoctl(WON_CTX *ctx,
				 uint32 ioctl,
				 void *inBuffer,
				 uint32 inLen)
{
	LINKED_LIST_ITEM *item;
	WON_MODULE_CTX *modctx;
	WON_IC ic;

	memset(&ic,0,sizeof(WON_IC));

	ic.inBuffer 	= inBuffer;
	ic.inLen	= inLen;

	item = listGetFirstItem(&ctx->moduleList);

	while (item)
	{
		modctx = (WON_MODULE_CTX *)item->value;

		if (modctx && modctx->moduleIoctl)
			modctx->moduleIoctl(modctx,ioctl,&ic);

		item = listGetNextItem(item);
	}

	return WON_STATUS_SUCCESS;
}

/* Scheduler Management */

WON_STATUS wonSchedulerRegister(WON_CTX *ctx,
				WON_MODULE_CTX *modctx,
				uint32 *schedulerId,
				uint32 timeFrame,
				uint32 executeTimes,
				uint32 (*func)(WON_SCHEDULED_TASK *task),
				void *schedulerContext)
{
	WON_SCHEDULED_TASK 	*task;
	uint32			newId = 0;

	if (!func)
		return WON_STATUS_FAILURE;

	while (1)
	{
		newId = rand();

		if (!listHasLongId(&ctx->schedulerList,newId))
			break;
	}

	task = (WON_SCHEDULED_TASK *)wonMemoryAllocate(sizeof(WON_SCHEDULED_TASK),1);

	task->id		= newId;
	if (schedulerId)
		*schedulerId	= newId;

	task->ctx		= ctx;
	task->modctx		= modctx;
	task->timeFrame		= timeFrame;
	task->executeTimes	= executeTimes;

	task->schedulerContext	= schedulerContext;

	task->callback		= func;

	listAddLongId(&ctx->schedulerList,newId,task);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonSchedulerDeregisterById(WON_CTX *ctx,
				  uint32 schedulerId)
{
	if (listDeleteLongId(&ctx->schedulerList,schedulerId))
		return WON_STATUS_SUCCESS;
	else
		return WON_STATUS_NOEXISTS;
}

WON_STATUS wonSchedulerDeregisterByCallback(WON_CTX *ctx,
					uint32 (*func)(WON_SCHEDULED_TASK *task))
{
	LINKED_LIST_ITEM *item;
	WON_SCHEDULED_TASK *task;
	WON_STATUS retval = WON_STATUS_FAILURE;

	item = listGetFirstItem(&ctx->schedulerList);

	while (item && item->value)
	{
		task = (WON_SCHEDULED_TASK *)item->value;

		if (task->callback == func)
		{
			listDeleteLongId(&ctx->schedulerList,task->id);

			retval = WON_STATUS_SUCCESS;

			break;
		}
	}

	return retval;
}

/* Low level handlers */

WON_STATUS wonLowlevelRecvRegisterLayeredHandler(WON_CTX *ctx,
						WON_MODULE_CTX *modctx,
						short layerPreference,
						uint32 (*func)(WON_MODULE_CTX *modctx,WON_PACKET *packet))
{
	if (listGetValueLongId(&ctx->lowlevelRecvHandlers,(uint32)func))
		return WON_STATUS_FAILURE;

	listAddLongId(&ctx->lowlevelRecvHandlers,(uint32)func,modctx);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonLowlevelRecvDeregisterLayeredHandler(WON_CTX *ctx,
						uint32 (*func)(WON_MODULE_CTX *modctx,WON_PACKET *packet))
{
	listDeleteLongId(&ctx->lowlevelRecvHandlers,(uint32)func);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonLowlevelSendRegisterLayeredHandler(WON_CTX *ctx,
						WON_MODULE_CTX *modctx,
						short layerPreference,
						uint32 (*func)(WON_MODULE_CTX *modctx,WON_PACKET *packet))
{
	if (listGetValueLongId(&ctx->lowlevelSendHandlers,(uint32)func))
		return WON_STATUS_FAILURE;

	listAddLongId(&ctx->lowlevelSendHandlers,(uint32)func,modctx);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonLowlevelSendDeregisterLayeredHandler(WON_CTX *ctx,
						uint32 (*func)(WON_MODULE_CTX *modctx,WON_PACKET *packet))
{
	listDeleteLongId(&ctx->lowlevelSendHandlers,(uint32)func);

	return WON_STATUS_SUCCESS;
}

/* Networking (IRC) */

WON_STATUS wonIrcSend(WON_CTX *ctx,
				WON_IRC_CONNECTION *irc,
				const char *fmt,
				...)
{
	char buf[2048];
	va_list ap;
	int len;

	buf[2047] = 0;

	va_start(ap,fmt);
	len = vsnprintf(buf,2047,fmt,ap);
	va_end(ap);

	if (ctx->lowlevelSendHandlers.items)
	{
		LINKED_LIST_ITEM *item;
		WON_IRC_PACKET ircp;
		WON_PACKET pkt;
		CACHE cache;

		pkt.type = WON_CONNECTION_IRC;
		ircp.ctx  = ctx;
		ircp.buffer = &cache;
		ircp.irc    = irc;
		pkt.pkt.irc = &ircp;

		cacheInitialize(&cache);

		strncpy(cache.buf,buf,len);
		cache.bufpos = len;

		cacheParseWords(&cache);
	
		for (item = listGetFirstItem(&ctx->lowlevelSendHandlers);
			item && item->value;
			item = listGetNextItem(item))
			if (((uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet))item->id.longId)((WON_MODULE_CTX *)item->value,&pkt) == WON_STATUS_FAILURE)
				break;

		cacheFree(&cache);

		if (item)
			return WON_STATUS_FAILURE;
	}

	write(irc->fd,buf,len);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonIrcEnum(WON_CTX *ctx,
				uint32 ircConnection,
				WON_IRC_CONNECTION **irc)
{
	LINKED_LIST_ITEM *item;
	uint32 x = 0;

	item = listGetFirstItem(&ctx->ircConnections);

	while (item)
	{
		if (x == ircConnection)
		{
			*irc = item->value;
			return WON_STATUS_SUCCESS;
		}

		x++;
	}

	return WON_STATUS_FAILURE;
}

WON_STATUS wonIrcCreateConnection(WON_CTX *ctx,
				unsigned long bindaddr,
				unsigned char *dst,
				uint32 port)
{
	WON_IRC_CONNECTION *irc;
	unsigned long resolveaddr;
	struct sockaddr_in info;
	int fd;

	if (!(fd = wonCreateTcpConnection(bindaddr,dst,port,&resolveaddr)))
		return WON_STATUS_FAILURE;

	info.sin_family	= AF_INET;
	info.sin_port	= htons(port);
	info.sin_addr.s_addr = resolveaddr;

	irc = wonMemoryAllocate(sizeof(WON_IRC_CONNECTION),1);

	irc->fd 	= fd;
	irc->addr	= resolveaddr;
	irc->port	= port;

	listAddLongId(&ctx->ircConnections,irc->fd,irc);

	wonConnectionAdd(ctx,WON_CONNECTION_IRC,irc->fd,&info);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonIrcDestroyConnection(WON_CTX *ctx,
				WON_IRC_CONNECTION *irc)
{
	wonModuleBroadcastIoctl(ctx,WON_IOCTL_IRC_LINK_DOWN,irc,sizeof(WON_IRC_CONNECTION));

	close(irc->fd);

	wonConnectionDelete(ctx,WON_CONNECTION_IRC,irc->fd);

	listDeleteLongId(&ctx->ircConnections,irc->fd);

	return WON_STATUS_SUCCESS;
}

WON_IRC_CONNECTION *wonIrcGetConnectionByHandle(WON_CTX *ctx, int fd)
{
	return (WON_IRC_CONNECTION *)listGetValueLongId(&ctx->ircConnections,fd);
}

void wonIrcConnectionSetNickName(WON_CTX *ctx, WON_IRC_CONNECTION *irc, const char *nickname)
{
	strncpy(irc->nickname,nickname,WON_MAX_NICKNAME_SIZE-1);
}

/* Networking (DCC) 		*/

WON_STATUS wonDccRegisterConnection(WON_CTX *ctx,
				WON_DCC_CONNECTION *dcc)
{
	struct sockaddr_in info;

	info.sin_family	= AF_INET;
	info.sin_port	= htons(dcc->port);
	info.sin_addr.s_addr = dcc->addr;

	listAddLongId(&ctx->dccConnections,dcc->fd,dcc);
	
	wonConnectionAdd(ctx,WON_CONNECTION_DCC,dcc->fd,&info);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonDccDeregisterConnection(WON_CTX *ctx,
				WON_DCC_CONNECTION *dcc)
{
	wonModuleBroadcastIoctl(ctx,
				WON_IOCTL_DCC_LINK_DOWN,
				dcc,
				sizeof(WON_DCC_CONNECTION));

	wonConnectionDelete(ctx,WON_CONNECTION_DCC,dcc->fd);

	listDeleteLongId(&ctx->dccConnections,dcc->fd);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonDccSend(WON_CTX *ctx,
				WON_DCC_CONNECTION *dcc,
				const char *fmt,
				...)
{
	char buf[2048];
	va_list ap;
	int len;

	buf[2047] = 0;

	va_start(ap,fmt);
	len = vsnprintf(buf,2047,fmt,ap);
	va_end(ap);

	if (ctx->lowlevelSendHandlers.items)
	{
		LINKED_LIST_ITEM *item;
		WON_DCC_PACKET dccp;
		WON_PACKET pkt;
		CACHE cache;

		pkt.type = WON_CONNECTION_DCC;
		dccp.ctx  = ctx;
		dccp.buffer = &cache;
		dccp.dcc    = dcc;
		pkt.pkt.dcc = &dccp;

		cacheInitialize(&cache);

		strncpy(cache.buf,buf,len);
		cache.bufpos = len;

		cacheParseWords(&cache);
	
		for (item = listGetFirstItem(&ctx->lowlevelSendHandlers);
			item && item->value;
			item = listGetNextItem(item))
			if (((uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet))item->id.longId)((WON_MODULE_CTX *)item->value,&pkt) == WON_STATUS_FAILURE)
				break;

		cacheFree(&cache);

		if (item)
			return WON_STATUS_FAILURE;
	}


	write(dcc->fd,buf,len);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonDccEnum(WON_CTX *ctx,
				uint32 dccConnection,
				WON_DCC_CONNECTION **dcc)
{
	LINKED_LIST_ITEM *item;
	uint32 x = 0;

	for (item = listGetFirstItem(&ctx->dccConnections);
			item && item->value && x != dccConnection;
			item = listGetNextItem(item), x++);

	if (item && item->value)
		*dcc = (WON_DCC_CONNECTION *)item->value;
	else
		return WON_STATUS_FAILURE;

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonDccSendAll(WON_CTX *ctx,
				const char *fmt,
				...)
{
	WON_DCC_CONNECTION *conn;
	int x = 0;
	char buf[2048];
	va_list ap;
	int len;
	
	buf[2047] = 0;
	
	va_start(ap,fmt);
	len = vsnprintf(buf,2047,fmt,ap);
	va_end(ap);

	while (wonDccEnum(ctx,x++,&conn) != WON_STATUS_FAILURE)
	{
		if (ctx->lowlevelSendHandlers.items)
		{
			LINKED_LIST_ITEM *item;
			WON_DCC_PACKET dccp;
			WON_PACKET pkt;
			CACHE cache;
	
			pkt.type = WON_CONNECTION_DCC;
			dccp.ctx  = ctx;
			dccp.buffer = &cache;
			dccp.dcc    = conn;
			pkt.pkt.dcc = &dccp;
	
			cacheInitialize(&cache);
	
			strncpy(cache.buf,buf,len);
			cache.bufpos = len;
	
			cacheParseWords(&cache);
		
			for (item = listGetFirstItem(&ctx->lowlevelSendHandlers);
				item && item->value;
				item = listGetNextItem(item))
				if (((uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet))item->id.longId)((WON_MODULE_CTX *)item->value,&pkt) == WON_STATUS_FAILURE)
					break;
	
			cacheFree(&cache);
	
			if (item)
				return WON_STATUS_FAILURE;
		}

		write(conn->fd,buf,len);
	}

	return WON_STATUS_SUCCESS;
}

WON_DCC_CONNECTION *wonDccGetConnectionByHandle(WON_CTX *ctx,
					int fd)
{
	return (WON_DCC_CONNECTION *)listGetValueLongId(&ctx->dccConnections,fd);
}

/* Configuration 		*/

WON_STATUS wonConfigurationSave(WON_CTX *ctx)
{
	FILE *fd = fopen(ctx->configurationFile,"w");
	LINKED_LIST_ITEM *item;

	if (ctx->logFile[0])
		fprintf(fd,"log_file    %s\n",ctx->logFile);

	for (item = listGetFirstItem(&ctx->moduleList);
		item && item->value;
		item = listGetNextItem(item))
	{
		WON_MODULE_CTX *modctx = (WON_MODULE_CTX *)item->value;
		
		fprintf(fd,"load_module %s %s\n",modctx->modulePath,modctx->moduleName);
	}

	wonModuleBroadcastIoctl(ctx,WON_IOCTL_CONFIG_SAVE,fd,4);

	fclose(fd);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonConfigurationReload(WON_CTX *ctx)
{
	wonLoaderConfigurationParse(ctx);

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonRegisterLayeredHandler(HASH_GROUP *hash,
				WON_CTX *ctx,
				WON_MODULE_CTX *modctx,
				short layerPreference,
				const char *item,
				int position,
				void *func)
{
	WON_HANDLERS *handlers = hashStringKeyGetValue(hash,item);
	WON_REGISTERED_HANDLER *newHandler = NULL;

	if (handlers)
	{
		if (layerPreference == WON_LAYER_FIRST)
		{
			newHandler = handlers->firstHandler;
			handlers->firstHandler = wonMemoryAllocate(sizeof(WON_REGISTERED_HANDLER),1);
			handlers->firstHandler->next = newHandler;
			newHandler->prev = handlers->firstHandler;
			newHandler = handlers->firstHandler;
		}
		else
		{
			newHandler = handlers->lastHandler;
			handlers->lastHandler = wonMemoryAllocate(sizeof(WON_REGISTERED_HANDLER),1);
			handlers->lastHandler->prev = newHandler;
			newHandler->next = handlers->lastHandler;
			newHandler = handlers->lastHandler;
		}
	}
	else
	{
		handlers = wonMemoryAllocate(sizeof(WON_HANDLERS),1);

		handlers->item = strdup(item);
		
		newHandler = handlers->lastHandler = handlers->firstHandler = wonMemoryAllocate(sizeof(WON_REGISTERED_HANDLER),1);

		hashStringKeyAdd(hash,handlers->item,handlers);
	}

	newHandler->modctx 	= modctx;
	newHandler->function	= (uint32)func;
	newHandler->position	= position;

	return WON_STATUS_SUCCESS;
}

WON_STATUS wonDeregisterLayeredHandler(HASH_GROUP *hash,
				WON_CTX *ctx,
				WON_MODULE_CTX *modctx,
				const char *item,
				int position,
				void *func)
{
	WON_HANDLERS *handlers = hashStringKeyGetValue(hash,item);
	WON_REGISTERED_HANDLER *curr = NULL;

	if (!handlers)
		return WON_STATUS_FAILURE;

	curr = handlers->firstHandler;

	while (curr)
	{
		if (curr->function == (uint32)func)
			break;

		curr = curr->next;
	}

	if (!curr)
		return WON_STATUS_FAILURE;

	if (curr->prev)
		curr->prev->next = curr->next;
	if (curr->next)
		curr->next->prev = curr->prev;

	if (curr == handlers->firstHandler)
		handlers->firstHandler = (curr->next)?curr->next:curr->prev;
	if (curr == handlers->lastHandler)
		handlers->lastHandler = (curr->prev)?curr->prev:curr->next;

	if (!handlers->firstHandler)
		handlers->firstHandler = handlers->lastHandler;

	/* No handlers left, remove ourselves. */
	if (!handlers->lastHandler)
	{
		wonMemoryFree(handlers->item);
		hashStringKeyDelete(hash,item);
	}

	return WON_STATUS_SUCCESS;
}

/* Logging			*/

WON_STATUS wonLog(WON_CTX *ctx,
		WON_MODULE_CTX *modctx,
		const char *fmt,
		...)
{
	struct tm *t;
	time_t lt;
	char buf[1024];
	va_list ap;

	if (!ctx->logfd)
		return WON_STATUS_FAILURE;

	lt = time(NULL);
	t = localtime(&lt);

	buf[1023] = 0;

	va_start(ap,fmt);
	vsnprintf(buf,1023,fmt,ap);
	va_end(ap);

	fprintf(ctx->logfd,"[%02i/%02i/%02i %02i:%02i:%02i]%s: %s\n",t->tm_mon+1,t->tm_mday,t->tm_year+1900,t->tm_hour,t->tm_min,t->tm_sec,(modctx)?modctx->moduleName:"",buf);
	fflush(ctx->logfd);

	return WON_STATUS_SUCCESS;
}

void wonConnectionAdd(WON_CTX *ctx,
			unsigned short type,
			int fd,
			struct sockaddr_in *s)
{
	ctx->wonConnections[ctx->numConnections].type	= type;
	ctx->wonConnections[ctx->numConnections].fd	= fd;
	memcpy(&ctx->wonConnections[ctx->numConnections].addr,s,sizeof(struct sockaddr_in));
	cacheInitialize(&ctx->wonConnections[ctx->numConnections].buffer);

	ctx->numConnections++;

	if (fd > ctx->highestFd)
		ctx->highestFd = fd;
}

void wonConnectionDelete(WON_CTX *ctx,
			 unsigned short type,
			 int fd)
{
	int x = 0, highestFd = 0;

	close(fd);

	for (; x < ctx->numConnections; x++)
	{
		if ((ctx->wonConnections[x].type == type) &&
		    (ctx->wonConnections[x].fd == fd))
			break;
	}

	cacheFree(&ctx->wonConnections[x].buffer);

	for (; x < ctx->numConnections; x++)
	{
		if (x+1 < WON_MAX_CONNECTIONS)
		{
			ctx->wonConnections[x].type = ctx->wonConnections[x+1].type;
			ctx->wonConnections[x].fd = ctx->wonConnections[x+1].fd;
			ctx->wonConnections[x].state = ctx->wonConnections[x+1].state;
			memcpy(&ctx->wonConnections[x].addr,&ctx->wonConnections[x+1].addr,sizeof(struct sockaddr_in));
		}
	}

	ctx->wonConnections[--ctx->numConnections].type = 0;
	ctx->wonConnections[ctx->numConnections].fd 	= 0;
	ctx->wonConnections[ctx->numConnections].state 	= 0;
	memset(&ctx->wonConnections[x].addr,0,sizeof(struct sockaddr_in));

	for (x = 0; x < ctx->numConnections; x++)
	{
		if (ctx->wonConnections[x].fd > highestFd)
			highestFd = ctx->wonConnections[x].fd;
	}

	ctx->highestFd = highestFd;
}

int wonCreateTcpConnection(unsigned long bindaddr,
				const char *dst,
				unsigned int port,
				unsigned long *resolveaddr)
{
	struct sockaddr_in s;
	int fd = 0;
	int ret = 0;

	if ((fd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) <= 0)
		return WON_STATUS_FAILURE;

	fcntl(fd,F_GETFL,&ret);
	ret |= O_NONBLOCK;
	fcntl(fd,F_SETFL,ret);

	s.sin_family	  = AF_INET;
	s.sin_port	  = 0;
	s.sin_addr.s_addr = bindaddr;

	bind(fd,&s,sizeof(s));

	s.sin_family 	= AF_INET;
	s.sin_port	= htons(port);
	s.sin_addr.s_addr = inet_addr(dst);

	if (s.sin_addr.s_addr == -1)
	{
		struct hostent *h;

		if (!(h = gethostbyname(dst)))
		{
			close(fd);
			return 0;
		}

		memcpy(&s.sin_addr.s_addr,h->h_addr,h->h_length);
	}

	if (resolveaddr)
		*resolveaddr = s.sin_addr.s_addr;

	if (connect(fd,(struct sockaddr *)&s,sizeof(s)) < 0)
	{
		if ((errno != EISCONN) && (errno != EAGAIN) && (errno != EINPROGRESS))
		{
			close(fd);	
			return 0;
		}
	}

	return fd;
}
