/*
 *              http://www.afro-productions.com
 * sc   12/2001         mmiller@hick.org        Matt Miller
 *
 *                          SLAPChat
 *            Secure Layer (Afro Production) Chat
 *                       (c) 2001-2002
 *			All rights reserved.
 *
 * All code below and above is released freely to the public with only two
 * limitations:
 *      a) Any desired use of this code for a for-profit nature requires express
 *         permission from the author.
 *      b) This code can be used to serve whatever purpose is desired.  With
 *         that in mind, the original author cannot be held liable for what this
 *         code is used for and therefore will NOT be held liable for the actions
 *         of others.  As always, give credit where credit is due.
 *
 * Further information about licensing and liability can be found in the file
 * 'COVER' that should have been included with this file.  If it was not,
 * please contact afro productions and request a copy.
 *
 ***** common library -- Common objects and functions definitions ******
 *
 * hash.c: generic hashing API that supports string and integer keys.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "bitwise.h"
#include "hash.h"

/*
 * hashInitialize
 * 	Initializes the hash group with its flags as well as allocating the
 * 	hash array and initializing the has mutex if required.
 *
 * 	Arguments (3):
 *		HASH_GROUP *hg		-- The hash group being initialized
 *		ulong flags		-- The flags to initialized the hash
 *					   group with
 *		uint hashBuckets	-- The number of buckets that the hash
 *					   should have.  This parameter should
 *					   be 0 if the hash group is
 *					   initialized with the flag of
 *					   HASH_FLAG_ISSTATIC.  This creates a
 *					   hash with 8192 buckets.
 */

void hashInitialize(HASH_GROUP *hg, unsigned long flags, unsigned int hashBuckets)
{
	hg->flags = flags;

	/* If the hash is static, set the number of buckets to the default
	 * number of hash buckets (8192) 
	 */
	if (IS_SET(flags,HASH_FLAG_ISSTATIC))
		hashBuckets = DEFAULT_HASH_BUCKETS;

	hg->hashBuckets = hashBuckets;
	hg->hashTable = (HASH_TABLE *)malloc(sizeof(HASH_TABLE) * hashBuckets);
}

/*
 * hashDeinitialize
 * 	Deinitializes the hash group by deallocating the hash array and
 * 	freeing the hash mutex if appropriate
 */

void hashDeinitialize(HASH_GROUP *hg)
{
	free(hg->hashTable);
}

/*
 * hashStringKeyAdd
 * 	Add a value to the hash with a string key
 *
 * 	Arguments (4):
 *		HASH_GROUP *hg		-- The hash group to add it to.
 *		const char *key		-- The key to hash with.
 *		void *value		-- The value to store in reference to
 *					   the key.
 *		short lock		-- Wether or not to lock the hash.
 */

int hashStringKeyAdd(HASH_GROUP *hg, const char *key, void *value)
{
	HASH_ENTRY *new = NULL;

	/* Determine if an entry exists */
	new = _hashStringKeyGetEntry(hg, key);

	/* If not, calculate the hash position and create a new entry */
	if (!new)
	{
		unsigned int id = 0;
		int hashPosition =_hashStringKeyGetPos(hg,key,&id,0);

		new = (HASH_ENTRY *)malloc(sizeof(HASH_ENTRY));

		new->next = hg->hashTable[hashPosition].hashEntry;
		hg->hashTable[hashPosition].hashEntry = new;

		new->id = id;
	}

	new->ptr = value;


	/* As long as the object being stored in the hash is required to be a
	 * generic object and the object itself is valid, increment the number
	 * of references
	 */

	return 1;
}

/*
 * hashStringKeyDelete
 * 	deletes an entry from a hash that is keyed with a string key
 *
 * 	Arguments (3):
 * 		HASH_GROUP *hg		-- The hash group being manipulated
 * 		const char *key		-- The key to delete from the hash.
 * 		short lock		-- Whether or not to lock the hash.
 */

int hashStringKeyDelete(HASH_GROUP *hg, const char *key)
{
	unsigned int id, retval = 0;
	int hashPosition;
	HASH_ENTRY *curr, *prev = NULL;

	/* Determine the hash position based off the key and get the first
	 * entry at that bucket. 
	 */
	hashPosition = _hashStringKeyGetPos(hg,key,&id,0);
	curr = hg->hashTable[hashPosition].hashEntry;

	/*
	 * Loop through the entries in that bucket and break out when a
	 * matching id is found.
	 */
	while (curr)
	{
		if (curr->id == id)
			break;

		prev = curr;
		curr = curr->next;
	}

	/* If a match was found, remove it from the list of entries at that
	 * bucket and destroy the value and entry.
	 */
	if (curr)
	{
		if (!prev)
			hg->hashTable[hashPosition].hashEntry = curr->next;
		else
			prev->next = curr->next;
	
		if ((curr->ptr) && (IS_SET(hg->flags, HASH_FLAG_FREEVALUE)))
			free(curr->ptr);

		free(curr);

		retval = 1;
	}

	return retval;
}

/*
 * hashStringKeyGetValue
 * 	get the value that is stored in the hash with key
 *
 * 	Arguments (3):
 * 		HASH_GROUP *hg		-- The hash group being manipulated.
 * 		const char *key		-- The key to use.
 * 		short stayLocked	-- Whether or not to keep the hash
 * 					   locked after locating the value.
 * 					   This is useful for times where you
 * 					   want to determine if a value is in
 * 					   the hash, and, if not,
 * 					   thread-safely insert a unique entry
 * 					   for that key into the hash.
 */

void *hashStringKeyGetValue(HASH_GROUP *hg, const char *key)
{
	HASH_ENTRY *val;
	void *value = NULL;

	val = _hashStringKeyGetEntry(hg,key);

	if (val)
		value = val->ptr;

	return value;
}

/*
 * _hashStringKeyGetPos
 * 	this is a local function that determines which bucket a key should be
 * 	stored in given the maximum number of buckets.
 *
 * 	Arguments (4):
 * 		HASH_GROUP *hg		-- The hash group being manipulated.
 * 		const char *key		-- The key to hash.
 * 		unsigned int *id	-- The calculated unique identifier
 * 					   for the string key.
 *		int dmax		-- The maximum length of the string to
 *					   use.  Default is 30.
 */

unsigned int _hashStringKeyGetPos(HASH_GROUP *hg, const char *key, unsigned int *id, int dmax)
{
	unsigned int c = 0;
	int max = (dmax <= 0)?30:dmax;

	/* While the key is valid and we haven't reached the maximum number of
	 * letters to hash
	 */
	while (*key && --max)
	{
		/* shift c one bit to the left and add the dec value of
		 * tolower of the current letter of the key.  set this
		 * operation to c.
		 */
		//c = (c << 1) + tolower(*key);
		c = (c << 2);
		c += tolower(*key);

		key++;
	}

	/* If the sender wants the id, set the value of id to c */
	if (id)
		*id = c;

	/* bit wrap c to the number of hashBuckets minus one */
	return (c & (hg->hashBuckets-1));
}

/*
 * _hashStringKeyGetEntry
 * 	gets the specific entry for a given key.
 *
 * 	Arguments (2):
 * 		HASH_GROUP *hg		-- The hash group being manipulated.
 * 		const char *key		-- The key to map with.
 *
 *  note: This function is always called in locked context if the hash group
 *        is thread safe.
 */

HASH_ENTRY *_hashStringKeyGetEntry(HASH_GROUP *hg, const char *key)
{
	unsigned int id;
	int hashPosition = _hashStringKeyGetPos(hg, key, &id, 0);
	HASH_ENTRY *curr;

	/* Get the root node of entries for the hash position of the key */
	curr = hg->hashTable[hashPosition].hashEntry;

	while (curr)
	{
		/* If the id's match, break out. */
		if (curr->id == id)
			break;

		curr = curr->next;
	}

	return curr;
}

/*
 * hashIntegerKeyAdd
 * 	maps an integer key to a value
 *
 * 	Arguments (4):
 * 		HASH_GROUP *hg		-- The hash group being manipulated.
 * 		int key			-- The key to map with.
 * 		void *value		-- The value to store as referenced by
 * 					   key.
 *		lock			-- Whether or not to lock the hash.
 */

int hashIntegerKeyAdd(HASH_GROUP *hg, int key, void *value)
{
	HASH_ENTRY *new = NULL;

	/* Determine if an entry already exists for this key */
	new = _hashIntegerKeyGetEntry(hg, key);

	/* If not, create one. */
	if (!new)
	{
		unsigned int id = 0;
		int hashPosition =_hashIntegerKeyGetPos(hg,key,&id,0);

		new = (HASH_ENTRY *)malloc(sizeof(HASH_ENTRY));

		new->next = hg->hashTable[hashPosition].hashEntry;
		hg->hashTable[hashPosition].hashEntry = new;

		new->id = id;
	}

	new->ptr = value;

	return 1;
}

/*
 * hashIntegerKeyDelete
 * 	deletes an entry from a hash that is keyed with an integer key
 *
 * 	Arguments (3):
 * 		HASH_GROUP *hg		-- The hash group being manipulated
 * 		int key			-- The key to delete from the hash.
 * 		short lock		-- Whether or not to lock the hash.
 */

int hashIntegerKeyDelete(HASH_GROUP *hg, int key)
{
	unsigned int id, retval = 0;
	int hashPosition;
	HASH_ENTRY *curr, *prev = NULL;

	/* Get the bucket position for that key */
	hashPosition = _hashIntegerKeyGetPos(hg,key,&id,0);

	curr = hg->hashTable[hashPosition].hashEntry;

	/* Loop through the list of entries at that bucket and find one with a
	 * matching id.
	 */
	while (curr)
	{
		if (curr->id == id)
			break;

		prev = curr;
		curr = curr->next;
	}

	/* If a match was found...remove it. */
	if (curr)
	{
		if (!prev)
			hg->hashTable[hashPosition].hashEntry = curr->next;
		else
			prev->next = curr->next;

		/* Destroy the value at which is mapped to by the key passed
		 * in.
		 */
		if ((curr->ptr) && (IS_SET(hg->flags, HASH_FLAG_FREEVALUE)))
			free(curr->ptr);

		free(curr);

		retval = 1;
	}

	return retval;
}

/*
 * hashIntegerKeyGetValue
 * 	get the value that is stored in the hash with key
 *
 * 	Arguments (3):
 * 		HASH_GROUP *hg		-- The hash group being manipulated.
 * 		int key			-- The key to use.
 * 		short stayLocked	-- Whether or not to keep the hash
 * 					   locked after locating the value.
 * 					   This is useful for times where you
 * 					   want to determine if a value is in
 * 					   the hash, and, if not,
 * 					   thread-safely insert a unique entry
 * 					   for that key into the hash.
 */

void *hashIntegerKeyGetValue(HASH_GROUP *hg, int key)
{
	HASH_ENTRY *val;
	void *value = NULL;

	val = _hashIntegerKeyGetEntry(hg,key);

	if (val)
		value = val->ptr;

	return value;
}

/* 
 * _hashIntegerKeyGetPos
 * 	determine the bucket of an integer key.
 *
 * 	Arguments (4):
 * 		HASH_GROUP *hg		-- The hash group being manipulated.
 * 		int key			-- The key to use.
 * 		unsigned int *id	-- The actual id of the key.
 * 		int dmax		-- not used.
 */

unsigned int _hashIntegerKeyGetPos(HASH_GROUP *hg, int key, unsigned int *id, int dmax)
{
	*id = key;

	/* Take the key modulus the number of buckets to determine which
	 * bucket to store the value in.
	 */

	return (key % hg->hashBuckets);
}

/*
 * _hashIntegerKeyGetEntry
 * 	get the actual entry of a given integer key.
 *
 * 	Arguments (2):
 * 		HASH_GROUP *hg		-- The hash group being manipulated.
 * 		int key			-- The key to get the entry of.
 *
 * note: This is always called in locked context if applicable.
 */

HASH_ENTRY *_hashIntegerKeyGetEntry(HASH_GROUP *hg, int key)
{
	unsigned int id;
	int hashPosition = _hashIntegerKeyGetPos(hg, key, &id, 0);
	HASH_ENTRY *curr;

	curr = hg->hashTable[hashPosition].hashEntry;

	while (curr)
	{
		if (curr->id == id)
			break;

		curr = curr->next;
	}

	return curr;
}
