/*
 *              http://www.afro-productions.com
 * sc   12/2001         mmiller@hick.org        Matt Miller
 *
 *                          SLAPChat
 *            Secure Layer (Afro Production) Chat
 *                       (c) 2001-2002
 *			All rights reserved.
 *
 * All code below and above is released freely to the public with only two
 * limitations:
 *      a) Any desired use of this code for a for-profit nature requires express
 *         permission from the author.
 *      b) This code can be used to serve whatever purpose is desired.  With
 *         that in mind, the original author cannot be held liable for what this
 *         code is used for and therefore will NOT be held liable for the actions
 *         of others.  As always, give credit where credit is due.
 *
 * Further information about licensing and liability can be found in the file
 * 'COVER' that should have been included with this file.  If it was not,
 * please contact afro productions and request a copy.
 *
 ***** common library -- Common objects and functions definitions ******
 *
 * list.c: generic list API that supports both unique and non-unique
 * identifiers of type long and string.
 *
 */

#include <stdlib.h>

#include "bitwise.h"
#include "list.h"

int listInitialize(LINKED_LIST *list,unsigned int flags)
{
	list->flags = flags;

	list->items = 0;
	list->list = NULL;

	return 1;
}

int listDeinitialize(LINKED_LIST *list)
{
	listFlush(list);

	return 1;
}

void listFlush(LINKED_LIST *list)
{
	LINKED_LIST_ITEM *item;

	item = list->list;

	while (list->list) 
	{
		item = list->list->next;

		if ((list->list->value) &&
			 (IS_SET(list->flags,LIST_FLAG_FREEVALUE)))
			 free(list->list->value);
	
		free(list->list);

		list->list = item;

		list->items--;
	}
}

inline LINKED_LIST_ITEM *listAddItem(LINKED_LIST *list, void *value)
{
	LINKED_LIST_ITEM *item;

	/* Add a reference now */

	if (!IS_SET(list->flags,LIST_FLAG_APPENDEND))
	{
		item = list->list;

		list->list = malloc(sizeof(LINKED_LIST_ITEM));
		memset(list->list, 0, sizeof(LINKED_LIST_ITEM));
	
		list->list->value 	= (void * *)value;
	
		if (item)
			item->prev	= list->list;
	
		list->list->next 	= item;
	
		list->items++;

		return list->list;
	}
	else
	{
		LINKED_LIST_ITEM *prev = NULL;
		
		item = list->list;

		if (!list->list)
			item = list->list = malloc(sizeof(LINKED_LIST_ITEM));
		else
		{
			while (item->next)
				item = item->next;
		
			item->next = malloc(sizeof(LINKED_LIST_ITEM));
			prev = item;
			item = item->next;
		}

		if (prev)
			prev->next = item;

		item->value = value;
		item->prev = prev;
		item->next = NULL;

		list->items++;

		return item;
	}
}

inline int listDeleteItem(LINKED_LIST *list, LINKED_LIST_ITEM *item)
{
	if (!item)
		return 0;

	if (item->prev)
		item->prev->next = item->next;
	else
		list->list	 = item->next;

	if (item->next)
		item->next->prev = item->prev;

	if ((list->list) && (list->list->value) &&
		 (IS_SET(list->flags,LIST_FLAG_FREEVALUE)))
		free(item->value);

	free(item);

	list->items--;

	return 1;
}

int listAddStringId(LINKED_LIST *list, const char *key, void *value)
{
	LINKED_LIST_ITEM *item = NULL;

	item = listAddItem(list, value);

	if (item)
		item->id.stringId = key;


	return 1;
}

int listAddUniqueStringId(LINKED_LIST *list, const char *key, void *value)
{
	LINKED_LIST_ITEM *item = NULL;

	if (listGetItemStringId(list,key))
		return 0;

	item = listAddItem(list, value);

	if (item)
		item->id.stringId = key;


	return 1;
}

int listDeleteStringId(LINKED_LIST *list, const char *key)
{
	int retval;

	retval = listDeleteItem(list,listGetItemStringId(list, key));


	return retval;
}

LINKED_LIST_ITEM *listGetItemStringId(LINKED_LIST *list, const char *key)
{
	LINKED_LIST_ITEM *item;

	item = list->list;

	while (item)
	{
		if ((item->id.stringId) && (!strcmp(item->id.stringId,key)))
			break;

		item = item->next;
	}

	return item;
}

void *listGetValueStringId(LINKED_LIST *list, const char *key)
{
	LINKED_LIST_ITEM *item;

	item = listGetItemStringId(list, key);

	return (item) ? item->value : NULL;
}

int listAddLongId(LINKED_LIST *list, long key, void *value)
{
	LINKED_LIST_ITEM *item;

	item = listAddItem(list, value);

	if (item)
		item->id.longId = key;


	return 1;
}

int listAddUniqueLongId(LINKED_LIST *list, long key, void *value)
{
	LINKED_LIST_ITEM *item;

	if (listGetItemLongId(list,key))
		return 0;

	item = listAddItem(list,value);

	if (item)
		item->id.longId = key;

	return 1;
}

int listDeleteLongId(LINKED_LIST *list, long key)
{
	int retval;

	retval = listDeleteItem(list,listGetItemLongId(list, key));

	return retval;
}

LINKED_LIST_ITEM *listGetItemLongId(LINKED_LIST *list, long key)
{
	LINKED_LIST_ITEM *item;

	item = list->list;

	while (item)
	{
		if (item->id.longId == key)
			break;

		item = item->next;
	}

	return item;
}

void *listGetValueLongId(LINKED_LIST *list, long key)
{
	LINKED_LIST_ITEM *item;

	item = listGetItemLongId(list,key);

	return (item) ? item->value : NULL; 
}

int listAddStringAndLongId(LINKED_LIST *list, const char *stringId, long longId, void *value)
{
	LINKED_LIST_ITEM *item;

	item = listAddItem(list, value);

	if (item)
	{
		item->id.stringId 	= stringId;
		item->id.longId 	= longId;
	}

	return 1;
}

int listAddUniqueStringAndLongId(LINKED_LIST *list, const char *stringId, long longId, void *value)
{
	LINKED_LIST_ITEM *item;

	if (listGetItemStringAndLongId(list, stringId, longId))
		return 0;

	item = listAddItem(list,value);

	if (item)
	{
		item->id.stringId 	= stringId;
		item->id.longId		= longId;
	}

	return 1;
}

int listDeleteStringAndLongId(LINKED_LIST *list, const char *stringId, long longId)
{
	int retval;

	retval = listDeleteItem(list, listGetItemStringAndLongId(list, stringId, longId));

	return retval;
}

LINKED_LIST_ITEM *listGetItemStringAndLongId(LINKED_LIST *list, const char *stringId, long longId)
{
	LINKED_LIST_ITEM *item = list->list;

	while (item)
	{
		if ((item->id.longId == longId) &&
		    ((stringId) && (!strcmp(stringId,item->id.stringId))))
			break;

		item = item->next;
	}

	return item;
}

void *listGetValueStringAndLongId(LINKED_LIST *list, const char *stringId, long longId)
{
	LINKED_LIST_ITEM *item;

	item = listGetItemStringAndLongId(list, stringId, longId);

	return (item) ? item->value : NULL;
}

LINKED_LIST_ITEM *listGetFirstItem(LINKED_LIST *list)
{
	return list->list;
}

LINKED_LIST_ITEM *listGetPreviousItem(LINKED_LIST_ITEM *item)
{
	return item->prev;
}

LINKED_LIST_ITEM *listGetNextItem(LINKED_LIST_ITEM *item)
{
	return item->next;
}

void *listGetItemValue(LINKED_LIST_ITEM *item)
{
	return ((item) && (item->value)) ? item->value : NULL;
}

short listHasStringId(LINKED_LIST *list, const char *key)
{
	LINKED_LIST_ITEM *item;
	short retval = 0;

	item = listGetFirstItem(list);

	while (item)
	{
		if ((item->value) && (item->id.stringId) && (!strcmp(key,item->id.stringId)))
		{
			retval = 1;
			break;
		}

		item = listGetNextItem(item);
	}


	return retval;
}

short listHasLongId(LINKED_LIST *list, long key)
{
	LINKED_LIST_ITEM *item;
	short retval = 0;

	item = listGetFirstItem(list);

	while (item)
	{
		if (item->id.longId == key)
		{
			retval = 1;
			break;
		}

		item = listGetNextItem(item);
	}


	return retval;
}

