#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "won.h"
#include "wonkrnl.h"

struct  {
	const char 	*str;
	uint32 		bit;
} access_relation[] = {
	{ "OPS",		WONKRNL_ACCESS_OPS 		},
	{ "AUTO_OPS",		WONKRNL_ACCESS_AUTO_OPS 	},
	{ "MASTER",		WONKRNL_ACCESS_MASTER		},
	{ "INVITES",		WONKRNL_ACCESS_INVITES		},
	{ "KICKS",		WONKRNL_ACCESS_KICKS		},
	{ "BANS",		WONKRNL_ACCESS_BANS		},
	{ "MSGS",		WONKRNL_ACCESS_MSGS		},
	{ NULL,			0				},
};

void configSave(WON_MODULE_CTX *modctx, FILE *fd)
{
	WONKRNL_MODULE_EXTENSION *krnl = WONKRNL_EXT(modctx->moduleExtension);
	LINKED_LIST_ITEM *networkItem, *channelItem, *serverItem, *userItem, *banItem, *hostItem;
	WONKRNL_NETWORK *network;

	for (networkItem = listGetFirstItem(&krnl->networks);
		networkItem && networkItem->value;
		networkItem = listGetNextItem(networkItem))
	{
		network = (WONKRNL_NETWORK *)networkItem->value;

		fprintf(fd,"start %s\n",network->name);

		fprintf(fd,"\tnickname        %s\n",network->nickname);
		fprintf(fd,"\tvhost           %hi.%hi.%hi.%hi\n",PRINT_IP(network->vhost));
		fprintf(fd,"\tdcc_user_prompt %s\n",network->dccUserPrompt);
		fprintf(fd,"\tdcc_pass_prompt %s\n",network->dccPassPrompt);
		fprintf(fd,"\tversion_reply   %s\n",network->versionReply);
		if (network->motdFile)
			fprintf(fd,"\tmotd_file       %s\n",network->motdFile);

		for (serverItem = listGetFirstItem(&network->servers);
			serverItem && serverItem->value;
			serverItem = listGetNextItem(serverItem))
		{
			WONKRNL_SERVER *server = (WONKRNL_SERVER *)serverItem->value;

			fprintf(fd,"\tserver          %hi.%hi.%hi.%hi %i\n",PRINT_IP(server->addr),server->port);
		}

		for (channelItem = listGetFirstItem(&network->channels);
			channelItem && channelItem->value;
			channelItem = listGetNextItem(channelItem))
		{
			WONKRNL_CHANNEL *channel = (WONKRNL_CHANNEL *)channelItem->value;

			fprintf(fd,"\tbegin_channel %s\n",channel->channel);

			if (channel->key)
				fprintf(fd,"\t\tchannel_key %s\n",channel->key);

			for (banItem = listGetFirstItem(&channel->bans);
				banItem && banItem->value;
				banItem = listGetNextItem(banItem))
			{
				WONKRNL_BAN *ban = (WONKRNL_BAN *)banItem->value;

				fprintf(fd,"\t\tadd_ban     %s %s\n",ban->mask,ban->reason);
			}

			fprintf(fd,"\tend_channel\n");
		}

		for (userItem = listGetFirstItem(&network->users);
			userItem && userItem->value;
			userItem = listGetNextItem(userItem))
		{
			WONKRNL_USER *user = (WONKRNL_USER *)userItem->value;
			int x;

			fprintf(fd,"\tbegin_user %s\n",user->username);
			
			fprintf(fd,"\t\tpassword       ");
			for (x = 0; x < 16; x++)
				fprintf(fd,"%02x",user->password[x]);
			fprintf(fd,"\n");

			fprintf(fd,"\t\tglobal_access  ");
			for (x = 0; access_relation[x].str; x++)
			{
				if (IS_SET(user->gaccess,access_relation[x].bit))
					fprintf(fd,"%s ",access_relation[x].str);
			}
			fprintf(fd,"\n");

			if (user->hosts.items)
			{
				for (hostItem = listGetFirstItem(&user->hosts);
					hostItem && hostItem->value;
					hostItem = listGetNextItem(hostItem))
					fprintf(fd,"\t\tallow_host     %s\n",(char *)hostItem->value);
			}

			if (user->channels.items)
			{
				for (channelItem = listGetFirstItem(&user->channels);
					channelItem && channelItem->value;
					channelItem = listGetNextItem(channelItem))
				{
					WONKRNL_USER_CHANNEL *userChannel = (WONKRNL_USER_CHANNEL *)channelItem->value;

					fprintf(fd,"\t\tchannel_access %s ",userChannel->channel->channel);
					
					if (userChannel->access == 0)
						fprintf(fd,"NONE\n");
					else
					{
						for (x = 0; access_relation[x].str; x++)
						{
							if (IS_SET(userChannel->access,access_relation[x].bit))
								fprintf(fd,"%s ",access_relation[x].str);
						}
						fprintf(fd,"\n");
					}
				}
			}

			fprintf(fd,"\tend_user\n");
		}

		fprintf(fd,"end\n");
	}
}

WONKRNL_NETWORK *configVerifySane(WON_CONFIG_PACKET *wcp, int args)
{
	if (!WONKRNL_EXT(wcp->modctx->moduleExtension)->currentNetwork)
	{
		fprintf(stdout,"wonkrnl: %lu: occured outside of network block\n",wcp->lineNumber);
		return NULL;
	}

	if (wcp->buffer->words < args)
	{	
		fprintf(stdout,"wonkrnl: %lu: missing arguments\n",wcp->lineNumber);
		return NULL; 
	}

	return (WONKRNL_NETWORK *)WONKRNL_EXT(wcp->modctx->moduleExtension)->currentNetwork;
}

uint32 configStart(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *network = NULL;
	WONKRNL_MODULE_EXTENSION *krnl = WONKRNL_EXT(wcp->modctx->moduleExtension);

	if (wcp->buffer->words < 2)
		return WON_STATUS_FAILURE;

	if (listHasStringId(&krnl->networks,wcp->buffer->str[1]))
	{
		fprintf(stdout,"wonkrnl: declaring network that is already declared '%s'.\n",wcp->buffer->str[1]);
		return WON_STATUS_FAILURE;
	}

	network = wonkrnlNetworkAdd(krnl,wcp->buffer->str[1]);

	krnl->currentNetwork = network;

	wonLog(wcp->ctx,wcp->modctx,"added network '%s'",wcp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 configEnd(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_MODULE_EXTENSION *krnl = WONKRNL_EXT(wcp->modctx->moduleExtension);
	WONKRNL_NETWORK *currNetwork = (WONKRNL_NETWORK *)krnl->currentNetwork;

	if (!currNetwork)
	{
		fprintf(stdout,"modkrnl: %lu: end of network declared but no start point defined.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	if (!currNetwork->nickname)
	{
		fprintf(stdout,"modkrnl: %lu: network '%s' declared without a nickname.\n",wcp->lineNumber,currNetwork->name);
		return WON_STATUS_FAILURE;
	}

	currNetwork = NULL;

	return WON_STATUS_SUCCESS;
}

uint32 configNickname(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	if (currNetwork->nickname);
		wonMemoryFree(currNetwork->nickname);
		
	currNetwork->nickname = strdup(wcp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 configVhost(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currNetwork->vhost	= inet_addr(wcp->buffer->str[1]);

	if (currNetwork->vhost == -1)
	{
		struct hostent *h;

		if (!(h = gethostbyname(wcp->buffer->str[1])))
		{
			fprintf(stdout,"wonkrnl: %lu: could not resolve vhost '%s'\n",wcp->lineNumber,wcp->buffer->str[1]);
			currNetwork->vhost = 0;
			return WON_STATUS_SUCCESS;
		}

		memcpy(&currNetwork->vhost,h->h_addr,h->h_length);
	}

	return WON_STATUS_SUCCESS;
}

uint32 configVersionReply(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	char fullBuffer[512];

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	if (currNetwork->versionReply)
		wonMemoryFree(currNetwork->versionReply);

	fullBuffer[511] = 0;
	cacheCompileWords(wcp->buffer,1,wcp->buffer->words," ",fullBuffer,511);

	currNetwork->versionReply = strdup(fullBuffer);

	return WON_STATUS_SUCCESS;
}

uint32 configDccUserPrompt(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	char fullBuffer[512];

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	if (currNetwork->dccUserPrompt)
		wonMemoryFree(currNetwork->dccUserPrompt);

	fullBuffer[511] = 0;
	cacheCompileWords(wcp->buffer,1,wcp->buffer->words," ",fullBuffer,511);

	currNetwork->dccUserPrompt = strdup(fullBuffer);

	return WON_STATUS_SUCCESS;
}

uint32 configDccPassPrompt(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	char fullBuffer[512];

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	if (currNetwork->dccPassPrompt)
		wonMemoryFree(currNetwork->dccPassPrompt);

	fullBuffer[511] = 0;
	cacheCompileWords(wcp->buffer,1,wcp->buffer->words," ",fullBuffer,511);

	currNetwork->dccPassPrompt = strdup(fullBuffer);

	return WON_STATUS_SUCCESS;
}

uint32 configMotdFile(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	char fullBuffer[512];

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	if (currNetwork->motdFile)
		wonMemoryFree(currNetwork->motdFile);

	fullBuffer[511] = 0;
	cacheCompileWords(wcp->buffer,1,wcp->buffer->words," ",fullBuffer,511);

	currNetwork->motdFile = strdup(fullBuffer);

	return WON_STATUS_SUCCESS;
}

uint32 configServer(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	wonkrnlServerAdd(currNetwork,wcp->buffer->str[1],(wcp->buffer->str[2])?atoi(wcp->buffer->str[2]):6667);

	return WON_STATUS_SUCCESS;
}

uint32 configBeginChannel(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	WONKRNL_CHANNEL *newChannel;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	if (listHasStringId(&currNetwork->channels,wcp->buffer->str[1]))
	{
		fprintf(stdout,"wonkrnl: %lu: a channel named '%s' already exists.\n",wcp->lineNumber,wcp->buffer->str[1]);
		return WON_STATUS_FAILURE;
	}

	newChannel = wonkrnlChannelAdd(currNetwork,wcp->buffer->str[1]);

	currNetwork->currentChannel = (void *)newChannel;

	return WON_STATUS_SUCCESS;
}

uint32 configEndChannel(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,1);

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currNetwork->currentChannel = NULL;

	return WON_STATUS_SUCCESS;
}

uint32 configChannelKey(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	WONKRNL_CHANNEL *currChannel;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currChannel = (WONKRNL_CHANNEL *)currNetwork->currentChannel;

	if (!currChannel)
	{
		fprintf(stdout,"wonkrnl: %lu: channel_key outside of channel block.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	if (currChannel->key)
		wonMemoryFree(currChannel->key);

	currChannel->key 	= strdup(wcp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 configAddBan(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	WONKRNL_CHANNEL *currChannel;
	WONKRNL_BAN	*newBan;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currChannel = (WONKRNL_CHANNEL *)currNetwork->currentChannel;

	if (!currChannel)
	{
		fprintf(stdout,"wonkrnl: %lu: add_ban outside of channel block.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	if (wcp->buffer->str[2])
	{
		char reason[512];

		reason[511] = 0;
		cacheCompileWords(wcp->buffer,2,wcp->buffer->words," ",reason,511);

		newBan = wonkrnlBanAdd(currChannel,wcp->buffer->str[1],reason);
	}
	else
		newBan = wonkrnlBanAdd(currChannel,wcp->buffer->str[1],"so filled with hate.");

	if (!newBan)
	{
		fprintf(stdout,"wonkrnl: %lu: a ban that matches this mask already exists.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	return WON_STATUS_SUCCESS;
}

WONKRNL_CHANNEL *configLookupChannel(WON_IRC_CONNECTION *irc, const char *channel)
{
	if (!irc || !irc->ircExtension)
		return NULL;

	return (WONKRNL_CHANNEL *)listGetValueStringId(&((WONKRNL_NETWORK *)irc->ircExtension)->channels,(char *)channel);
}

uint32 configBeginUser(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	WONKRNL_USER *newUser;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	if (listHasStringId(&currNetwork->users,wcp->buffer->str[1]))
	{
		fprintf(stdout,"wonkrnl: %lu: user %s already exists in network.\n",wcp->lineNumber,wcp->buffer->str[1]);
		return WON_STATUS_FAILURE;
	}

	newUser = wonkrnlUserAdd(currNetwork,wcp->buffer->str[1]);;

	currNetwork->currentUser= (void *)newUser;

	return WON_STATUS_SUCCESS;
}

uint32 configEndUser(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,1);

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currNetwork->currentUser = NULL;

	return WON_STATUS_SUCCESS;
}

uint32 configPassword(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	WONKRNL_USER *currUser;
	int len, i = 0, y = 0;
	char buf[3] = { 0, 0, 0};

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currUser = (WONKRNL_USER *)currNetwork->currentUser;

	if (!currUser)
	{
		fprintf(stdout,"wonkrnl: %lu: password outside of user block.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	len = strlen(wcp->buffer->str[1]);

	for (i = 0; i+1 < len && y < 16; i+=2, y++)
	{
		buf[0] = wcp->buffer->str[1][i]; buf[1] = wcp->buffer->str[1][i+1];

		currUser->password[y] = strtoul(buf,NULL,16);
	}

	return WON_STATUS_SUCCESS;
}

uint32 configGlobalAccess(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	WONKRNL_USER *currUser;
	int x, y;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currUser = (WONKRNL_USER *)currNetwork->currentUser;

	if (!currUser)
	{
		fprintf(stdout,"wonkrnl: %lu: global_access outside of user block.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	for (x = 1; x < wcp->buffer->words; x++)
	{
		for (y = 0; access_relation[y].str; y++)
		{
			if (!strcasecmp(wcp->buffer->str[x],access_relation[y].str))
				BIT_SET(currUser->gaccess,access_relation[y].bit);
		}
	}

	return WON_STATUS_SUCCESS;
}

uint32 configAllowHost(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,2);
	WONKRNL_USER *currUser;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currUser = (WONKRNL_USER *)currNetwork->currentUser;

	if (!currUser)
	{
		fprintf(stdout,"wonkrnl: %lu: allow_host outside of user block.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	wonkrnlUserHostAdd(currUser,wcp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 configChannelAccess(WON_CONFIG_PACKET *wcp)
{
	WONKRNL_NETWORK *currNetwork = configVerifySane(wcp,3);
	WONKRNL_USER *currUser;
	WONKRNL_USER_CHANNEL *userChannel = NULL;
	WONKRNL_CHANNEL *currChannel;
	int y, x, needAdd = 0;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	currUser = (WONKRNL_USER *)currNetwork->currentUser;
	currChannel = (WONKRNL_CHANNEL *)listGetValueStringId(&currNetwork->channels,wcp->buffer->str[1]);

	if (!currUser)
	{
		fprintf(stdout,"wonkrnl: %lu: channel_access outside of user block.\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	if (!currChannel)
	{
		fprintf(stdout,"wonkrnl: %lu: channel '%s' has not been defined yet.\n",wcp->lineNumber,wcp->buffer->str[1]);
		return WON_STATUS_FAILURE;
	}

	userChannel = wonkrnlUserChannelAdd(currUser,wcp->buffer->str[1]);

	if (!userChannel)
	{
		fprintf(stdout,"wonkrnl: %lu: channel perms already exist\n",wcp->lineNumber);
		return WON_STATUS_FAILURE;
	}

	for (x = 2; x < wcp->buffer->words; x++)
	{
		for (y = 0; access_relation[y].str; y++)
		{
			if (!strcasecmp(wcp->buffer->str[x],access_relation[y].str))
				BIT_SET(userChannel->access,access_relation[y].bit);
		}
	}

	if (needAdd)
		listAddStringId(&currUser->channels,userChannel->channel->channel,userChannel);

	return WON_STATUS_SUCCESS;
}

WONKRNL_USER *configLookupUser(WON_IRC_CONNECTION *irc,const char *user)
{
	if (!irc || !irc->ircExtension)
		return NULL;

	return (WONKRNL_USER *)listGetValueStringId(&((WONKRNL_NETWORK *)irc->ircExtension)->users,(char *)user);
}

WONKRNL_USER *configLookupUserByMask(WON_IRC_CONNECTION *irc,const char *mask)
{
	LINKED_LIST_ITEM *item, *hostItem;
	WONKRNL_USER *user;

	if (!irc || !irc->ircExtension)
		return NULL;

	for (item = listGetFirstItem(&((WONKRNL_NETWORK *)irc->ircExtension)->users);
			item && item->value;
			item = listGetNextItem(item))
	{
		user = (WONKRNL_USER *)item->value;
		for (hostItem = listGetFirstItem(&user->hosts);
				hostItem && hostItem->value;
				hostItem = listGetNextItem(hostItem))
		{
			if (wonWildcardMatch((unsigned char *)hostItem->value,(unsigned char *)mask))
				return user;
		}
	}

	return NULL;
}
