#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>

#include  "won.h"
#include "wonkrnl.h"

uint32 wonkrnlResolveAddress(const char *address)
{
	uint32 addr = 0;

	addr = inet_addr(address);
	
	if (addr == -1)
	{
		struct hostent *h;

		if (!(h = gethostbyname(address)))
			return -1;

		memcpy(&addr,h->h_addr,h->h_length);
	}

	return addr;
}

WONKRNL_NETWORK *wonkrnlNetworkAdd(WONKRNL_MODULE_EXTENSION *krnl,const char *network)
{
	WONKRNL_NETWORK *newNetwork;
	
	if (wonkrnlNetworkHas(krnl,network))
		return NULL;

	newNetwork = wonMemoryAllocate(sizeof(WONKRNL_NETWORK),1);

	newNetwork->name 	= strdup(network);
        newNetwork->versionReply   = strdup(WON_DEFAULT_VERSION_REPLY);
	newNetwork->dccUserPrompt  = strdup(WON_DEFAULT_DCC_USER_PROMPT);
	newNetwork->dccPassPrompt  = strdup(WON_DEFAULT_DCC_PASS_PROMPT);

	listInitialize(&newNetwork->servers,LIST_FLAG_FREEVALUE);
	listInitialize(&newNetwork->users,LIST_FLAG_FREEVALUE);

	listAddStringId(&krnl->networks,newNetwork->name,newNetwork);

	return newNetwork;
}

WONKRNL_NETWORK *wonkrnlNetworkHas(WONKRNL_MODULE_EXTENSION *krnl,const char *network)
{
	return (WONKRNL_NETWORK *)listGetValueStringId(&krnl->networks,(char *)network);
}

uint32 wonkrnlNetworkDelete(WONKRNL_MODULE_EXTENSION *krnl,const char *network)
{
	WONKRNL_NETWORK *currNetwork = wonkrnlNetworkHas(krnl,network);
	LINKED_LIST_ITEM *item, *banItem;
	WONKRNL_USER 	*user;
	WONKRNL_CHANNEL *chan;
	WONKRNL_BAN 	*ban;

	if (!currNetwork)
		return WON_STATUS_FAILURE;

	wonMemoryFree(currNetwork->name);
	wonMemoryFree(currNetwork->versionReply);
	wonMemoryFree(currNetwork->dccUserPrompt);
	wonMemoryFree(currNetwork->dccPassPrompt);

	listDeinitialize(&currNetwork->servers);	/* Free up servers 	*/

	for (item = listGetFirstItem(&currNetwork->users);
			item && item->value;
			item = listGetNextItem(item))
	{
		user = (WONKRNL_USER *)item->value;

		wonMemoryFree(user->username);

		listDeinitialize(&user->channels);
		listDeinitialize(&user->hosts);
	}

	for (item = listGetFirstItem(&currNetwork->channels);
			item && item->value;
			item = listGetNextItem(item))
	{
		chan = (WONKRNL_CHANNEL *)item->value;

		wonMemoryFree(chan->channel);

		for (banItem = listGetFirstItem(&chan->bans);
				banItem && banItem->value;
				banItem = listGetNextItem(banItem))
		{
			ban = (WONKRNL_BAN *)banItem->value;

			wonMemoryFree(ban->mask);
			wonMemoryFree(ban->reason);
		}

		listDeinitialize(&chan->bans);
	}

	listDeinitialize(&currNetwork->channels);

	return WON_STATUS_SUCCESS;
}

WONKRNL_SERVER *wonkrnlServerAdd(WONKRNL_NETWORK *network,const char *server,unsigned int port)
{
	WONKRNL_SERVER *newServer;

	if (wonkrnlServerHas(network,server,port))
		return NULL;
	
	newServer = (WONKRNL_SERVER *)wonMemoryAllocate(sizeof(WONKRNL_SERVER),1);

	newServer->network	= network;

	newServer->addr		= wonkrnlResolveAddress(server);
	newServer->port 	= port;

	if ((newServer->addr == -1) || (!listAddLongId(&network->servers,newServer->addr,newServer)))
	{
		wonMemoryFree(newServer);
		return NULL;
	}

	return newServer;
}

WONKRNL_SERVER *wonkrnlServerHas(WONKRNL_NETWORK *network,const char *server,unsigned int port)
{
	LINKED_LIST_ITEM *item;
	WONKRNL_SERVER *currServer;
	uint32 addr = wonkrnlResolveAddress(server);

	if (addr == -1)
		return NULL;

	for (item = listGetFirstItem(&network->servers);
		item && item->value;
		item = listGetNextItem(item))
	{
		currServer = (WONKRNL_SERVER *)item->value;

		if ((currServer->addr == addr) && 
			(currServer->port == port))
			return currServer;
	}

	return NULL;
}

uint32 wonkrnlServerDelete(WONKRNL_NETWORK *network,const char *server,unsigned int port)
{
	uint32 addr = wonkrnlResolveAddress(server);

	listDeleteLongId(&network->servers,addr);

	return WON_STATUS_SUCCESS;
}

WONKRNL_USER *wonkrnlUserAdd(WONKRNL_NETWORK *network,const char *username)
{
	WONKRNL_USER *user;

	if (wonkrnlUserHas(network,username))
		return NULL;
	
	user = (WONKRNL_USER *)wonMemoryAllocate(sizeof(WONKRNL_USER),1);

	user->username	= strdup(username);
	user->network	= network;

	listInitialize(&user->channels,LIST_FLAG_FREEVALUE);
	listInitialize(&user->hosts,LIST_FLAG_FREEVALUE);

	if (!listAddStringId(&network->users,user->username,user))
	{
		wonMemoryFree(user);
		return NULL;
	}

	return user;
}

WONKRNL_USER *wonkrnlUserHas(WONKRNL_NETWORK *network,const char *username)
{
	return (WONKRNL_USER *)listGetValueStringId(&network->users,username);
}

uint32 wonkrnlUserDelete(WONKRNL_NETWORK *network,const char *username)
{
	WONKRNL_USER *user = (WONKRNL_USER *)listGetValueStringId(&network->users,username);

	if (!user)
		return WON_STATUS_FAILURE;

	wonMemoryFree(user->username);

	listDeinitialize(&user->channels);
	listDeinitialize(&user->hosts);

	listDeleteStringId(&network->users,username);

	return WON_STATUS_SUCCESS;
}

WONKRNL_CHANNEL *wonkrnlChannelAdd(WONKRNL_NETWORK *network,const char *channel)
{
	WONKRNL_CHANNEL *newChannel;

	if (wonkrnlChannelHas(network,channel))
		return NULL;
	
	newChannel = (WONKRNL_CHANNEL *)wonMemoryAllocate(sizeof(WONKRNL_NETWORK),1);

	newChannel->channel 	= strdup(channel);

	listInitialize(&newChannel->bans,LIST_FLAG_FREEVALUE);

	listAddStringId(&network->channels,newChannel->channel,newChannel);

	return newChannel;
}

WONKRNL_CHANNEL *wonkrnlChannelHas(WONKRNL_NETWORK *network,const char *channel)
{
	return (WONKRNL_CHANNEL *)listGetValueStringId(&network->channels,channel);
}

uint32 wonkrnlChannelDelete(WONKRNL_NETWORK *network,const char *channel)
{
	WONKRNL_CHANNEL *newChannel = (WONKRNL_CHANNEL *)listGetValueStringId(&network->channels,channel);

	if (!newChannel)
		return WON_STATUS_FAILURE;

	wonMemoryFree(newChannel->channel);

	listDeinitialize(&newChannel->bans);

	listDeleteStringId(&network->channels,channel);

	return WON_STATUS_SUCCESS;
}

WONKRNL_BAN *wonkrnlBanAdd(WONKRNL_CHANNEL *channel,const char *mask,const char *reason)
{
	WONKRNL_BAN *newBan;

	if (wonkrnlBanHas(channel,mask))
		return NULL;

	newBan 	= (WONKRNL_BAN *)wonMemoryAllocate(sizeof(WONKRNL_BAN),1);

	newBan->mask	= strdup(mask);
	newBan->reason	= strdup(reason);
	newBan->channel	= channel;

	listAddStringId(&channel->bans,newBan->mask,newBan);

	return newBan;
}

WONKRNL_BAN *wonkrnlBanHas(WONKRNL_CHANNEL *channel,const char *mask)
{
	LINKED_LIST_ITEM *item;
	WONKRNL_BAN *ban;

	for (item = listGetFirstItem(&channel->bans);
		item && item->value;
		item = listGetNextItem(item))
	{
		ban = (WONKRNL_BAN *)item->value;

		if (wonWildcardMatch(ban->mask,(unsigned char *)mask))
			return ban;
	}

	return NULL;
}

uint32 wonkrnlBanDelete(WONKRNL_CHANNEL *channel,const char *mask)
{
	WONKRNL_BAN *ban;
	LINKED_LIST_ITEM *item;
	WON_STATUS retval = WON_STATUS_FAILURE;

	item = listGetFirstItem(&channel->bans);

	while (item && item->value)
	{
		ban = (WONKRNL_BAN *)item->value;

		if (wonWildcardMatch((unsigned char *)mask,ban->mask))
		{
			char *oldMask = ban->mask;
			wonMemoryFree(ban->reason);

			listDeleteStringId(&channel->bans,ban->mask);

			item = listGetFirstItem(&channel->bans);

			wonMemoryFree(oldMask);

			retval = WON_STATUS_SUCCESS;

			continue;
		}

		item = listGetNextItem(item);
	}

	return retval;
}

WONKRNL_USER_CHANNEL *wonkrnlUserChannelAdd(WONKRNL_USER *user,const char *channel)
{
	WONKRNL_USER_CHANNEL *newUserChannel;
	WONKRNL_CHANNEL *currChannel = wonkrnlChannelHas(user->network,channel);

	if ((wonkrnlUserChannelHas(user,channel)) ||
		(!currChannel))
		return NULL;

	newUserChannel = (WONKRNL_USER_CHANNEL *)wonMemoryAllocate(sizeof(WONKRNL_USER_CHANNEL),1);

	newUserChannel->user 	= user;
	newUserChannel->channel = currChannel;

	listAddStringId(&user->channels,currChannel->channel,newUserChannel);

	return newUserChannel;
}

WONKRNL_USER_CHANNEL *wonkrnlUserChannelHas(WONKRNL_USER *user,const char *channel)
{
	return (WONKRNL_USER_CHANNEL *)listGetValueStringId(&user->channels,(char *)channel);
}

uint32 wonkrnlUserChannelDelete(WONKRNL_USER *user,const char *channel)
{
	if (!listDeleteStringId(&user->channels,channel))
		return WON_STATUS_FAILURE;

	return WON_STATUS_SUCCESS;
}

const char *wonkrnlUserHostAdd(WONKRNL_USER *user,const char *host)
{
	char *newHost;

	if (wonkrnlUserHostHas(user,host))
		return NULL;

	newHost 	= strdup(host);

	listAddStringId(&user->hosts,newHost,newHost);

	return newHost;
}

const char *wonkrnlUserHostHas(WONKRNL_USER *user,const char *host)
{
	LINKED_LIST_ITEM *item;
	char *currHost;

	for (item = listGetFirstItem(&user->hosts);
			item && item->value;
			item = listGetNextItem(item))
	{
		currHost = (char *)item->value;

		if (wonWildcardMatch(currHost,(unsigned char *)host))
			return currHost;
	}

	return NULL;
}

uint32 wonkrnlUserHostDelete(WONKRNL_USER *user,const char *host)
{
	LINKED_LIST_ITEM *item = listGetFirstItem(&user->hosts);
	WON_STATUS retval = WON_STATUS_FAILURE;
	char *currHost;

	while (item && item->value)
	{
		currHost = (char *)item->value;

		if (wonWildcardMatch((unsigned char *)host,currHost))
		{
			retval = WON_STATUS_SUCCESS;
			listDeleteStringId(&user->hosts,currHost);
			item = listGetFirstItem(&user->hosts);
			continue;
		}

		item = listGetNextItem(item);
	}	

	return retval;
}
