#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#ifdef SUNOS
        #include <stropts.h>
        #define NREAD   I_NREAD
#else
        #define NREAD   FIONREAD
#endif
#include <string.h>
#include <errno.h>

#include "won.h"
#include "wonloader.h"

/* Configuration */

WON_STATUS wonLoaderProcessPerTypePacket(WON_CTX *ctx, WON_CONNECTION *conn);
void wonloaderProcessScheduledTasks(WON_CTX *ctx);

void wonLoaderConfigurationParse(WON_CTX *ctx)
{
	FILE *fd = fopen(ctx->configurationFile,"r");
	WON_CONFIG_PACKET configPacket;
	CACHE cache;

	if (!fd)
		wonLoaderExit(ctx,"failed to open configuration file '%s'",ctx->configurationFile);

	configPacket.ctx	= ctx;
	configPacket.modctx	= NULL;
	configPacket.buffer	= &cache;
	configPacket.lineNumber = 0;

	cacheInitialize(&cache);

	while (fgets(cache.buf,BUF_SIZE-1,fd))
	{
		configPacket.lineNumber++;
		cache.bufpos = strlen(cache.buf);

		cacheParseWords(&cache);

		if (cache.words)
		{
			WON_HANDLERS *handlers = hashStringKeyGetValue(&ctx->configHandlers,
									cache.str[0]);
			if (handlers)
			{
				WON_REGISTERED_HANDLER *curr = handlers->firstHandler;
				WON_STATUS status;

				while (curr)
				{
					configPacket.modctx = curr->modctx;
					if ((status = ((uint32 (*)(WON_CONFIG_PACKET *))(curr->function))(&configPacket)) != WON_STATUS_SUCCESS)
						wonLoaderExit(ctx,"line %i %s: aborting.",configPacket.lineNumber,cache.str[0]);

					if (status == WON_STATUS_CLAIMPACKET)
						break;

					curr = curr->next;
				}
			}
		}
	}

	cacheFree(&cache);

	fclose(fd);
}

uint32 wonLoaderConfigurationLoadModule(WON_CONFIG_PACKET *conf)
{
	if (conf->buffer->words < 3)
	{
		wonLog(conf->ctx,NULL,"load_module: syntax error line %lu\n",conf->lineNumber);
		return WON_STATUS_SUCCESS;
	}

	if (wonModuleLoad(conf->ctx,
			conf->buffer->str[1],
			conf->buffer->str[2]) != WON_STATUS_SUCCESS)
		fprintf(stdout,"wonloader: load_module: %s failed to load.\n",conf->buffer->str[2]);
	else
		fprintf(stdout,"wonloader: load_module: %s loaded.\n",conf->buffer->str[2]);

	return WON_STATUS_SUCCESS;
}

uint32 wonLoaderConfigurationLogFile(WON_CONFIG_PACKET *conf)
{
	if (conf->buffer->words < 2)
		return WON_STATUS_FAILURE;

	conf->ctx->logfd = fopen(conf->buffer->str[1],"a");
	strncpy(conf->ctx->logFile,conf->buffer->str[1],MAX_PATH-1);

	wonLog(conf->ctx,NULL,"Logging started.");

	return WON_STATUS_SUCCESS;
}

/* Execution 	*/

void wonLoaderExecute(WON_CTX *ctx)
{
	fd_set fdread;
	WON_STATUS retval, x, y;
	struct timeval tv;
	char readBuffer[BUF_SIZE];
	int readLen, remainder;

	wonModuleBroadcastIoctl(ctx,WON_IOCTL_LOAD_COMPLETE,NULL,0);

	while (1)
	{
		if (ctx->wantComplete)
			break;

		wonloaderProcessScheduledTasks(ctx);

		FD_ZERO(&fdread);
		for (x = 0; x < ctx->numConnections; x++)
			FD_SET(ctx->wonConnections[x].fd,&fdread);
		
		tv.tv_sec  = 1;
		tv.tv_usec = 0;

		retval = select(ctx->highestFd + 1, &fdread, NULL, NULL, &tv);

		if (!retval)
			continue;

		for (x = 0; x < ctx->numConnections && retval != WON_STATUS_CONNCLOSED; x++)
		{
			int readBytes = 1;
			if (!FD_ISSET(ctx->wonConnections[x].fd,&fdread))
				continue;

			if (!IS_SET(ctx->wonConnections[x].state,WON_SOCKET_STATE_CONNECTED))
			{
				if (connect(ctx->wonConnections[x].fd,
					(struct sockaddr *)&ctx->wonConnections[x].addr,
					sizeof(ctx->wonConnections[x].addr)) < 0)
				{
					if ((errno != EISCONN) &&
						(errno != EAGAIN) && 
						(errno != EINPROGRESS))
						readBytes = 0;
					else if (errno == EISCONN)
					{
						BIT_SET(ctx->wonConnections[x].state,WON_SOCKET_STATE_CONNECTED);

						if (ctx->wonConnections[x].type == WON_CONNECTION_IRC)
						{
							WON_IRC_CONNECTION *currConnection = wonIrcGetConnectionByHandle(ctx,ctx->wonConnections[x].fd);
							wonModuleBroadcastIoctl(ctx,WON_IOCTL_IRC_LINK_UP,currConnection,sizeof(WON_IRC_CONNECTION));
						}
						else if (ctx->wonConnections[x].type == WON_CONNECTION_DCC)
						{
							WON_DCC_CONNECTION *dccConnection = wonDccGetConnectionByHandle(ctx,ctx->wonConnections[x].fd);
							wonModuleBroadcastIoctl(ctx,WON_IOCTL_DCC_LINK_UP,dccConnection,sizeof(WON_DCC_CONNECTION));
						}
					}
				}
			}
		
			if (readBytes)
				ioctl(ctx->wonConnections[x].fd,NREAD,&readBytes);

			if (!readBytes)
			{
				switch (ctx->wonConnections[x].type)
				{
					case WON_CONNECTION_IRC:
						{
							WON_IRC_CONNECTION *currConnection = wonIrcGetConnectionByHandle(ctx,ctx->wonConnections[x].fd);
							wonIrcDestroyConnection(ctx,currConnection);
						}
						break;
					case WON_CONNECTION_DCC:
						{
							WON_DCC_CONNECTION *dccConnection = wonDccGetConnectionByHandle(ctx,ctx->wonConnections[x].fd);
							wonDccDeregisterConnection(ctx,dccConnection);
						}
						break;
				}
				
				retval = WON_STATUS_CONNCLOSED;

				continue;
			}

			remainder = (BUF_SIZE-1) - ctx->wonConnections[x].buffer.bufpos;

			while (readBytes > 0 && retval != WON_STATUS_CONNCLOSED)
			{
				readLen = read(ctx->wonConnections[x].fd,readBuffer,(readBytes > remainder)?remainder:readBytes);

				for (y = 0; y < readLen; y++)
				{
					if ((readBuffer[y] == '\n') || (ctx->wonConnections[x].buffer.bufpos >= BUF_SIZE-1))
					{
						ctx->wonConnections[x].buffer.buf[ctx->wonConnections[x].buffer.bufpos] = 0;
						cacheParseWords(&ctx->wonConnections[x].buffer);

						if ((retval = wonLoaderProcessPerTypePacket(ctx,&ctx->wonConnections[x])) == WON_STATUS_CONNCLOSED)
							break;

						memset(ctx->wonConnections[x].buffer.buf,0,ctx->wonConnections[x].buffer.bufpos);
						ctx->wonConnections[x].buffer.bufpos = 0;
					}
					else if (readBuffer[y] != '\n')
						ctx->wonConnections[x].buffer.buf[ctx->wonConnections[x].buffer.bufpos++] = readBuffer[y];
				}

				readBytes -= readLen;
			}
		}
	}
}
		
void wonloaderProcessScheduledTasks(WON_CTX *ctx)
{
	LINKED_LIST_ITEM *item;
	WON_SCHEDULED_TASK *task;
	uint32 currTime = time(NULL);

	for (item = listGetFirstItem(&ctx->schedulerList);
		item && item->value;
		item = listGetNextItem(item))
	{
		task = (WON_SCHEDULED_TASK *)item->value;

		if (task->timeFrame+task->lastExecute <= currTime)
		{
			if (task->executeTimes != -1)
				task->executeTimes--;

			task->callback(task);
			
			task->lastExecute = currTime;

			if (task->executeTimes <= 0)
			{
				wonSchedulerDeregisterById(ctx,task->id);
				break;
			}
		}
	}
}

WON_STATUS wonLoaderProcessPerTypePacket(WON_CTX *ctx, WON_CONNECTION *conn)
{
	LINKED_LIST_ITEM *item = NULL;
	WON_PACKET packet;
	WON_STATUS retval = WON_STATUS_SUCCESS;
	int x = 0;

	memset(&packet,0,sizeof(WON_PACKET));
	packet.type	= conn->type;

	switch (conn->type)
	{
		case WON_CONNECTION_IRC:
			{
				WON_IRC_PACKET wip;

				memset(&wip,0,sizeof(WON_IRC_PACKET));

				wip.ctx = ctx;
				wip.buffer = &conn->buffer;
				wip.irc    = wonIrcGetConnectionByHandle(ctx,conn->fd);

				packet.pkt.irc = &wip;

				for (item = listGetFirstItem(&ctx->lowlevelRecvHandlers);
						item && item->value;
						item = listGetNextItem(item))
					if (((uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet))item->id.longId)((WON_MODULE_CTX *)item->value,&packet) == WON_STATUS_FAILURE)
						break;

				if (item)
					break;

				if (conn->buffer.words >= 1)
				{
					if (wonWildcardMatch(":*!*@*",conn->buffer.str[0]))
					{
						int copySize = (strchr(conn->buffer.str[0],'!')) - (conn->buffer.str[0]+1);

						strncpy(wip.fromNick,
								conn->buffer.str[0]+1,
								(copySize > WON_MAX_NICKNAME_SIZE-1)?WON_MAX_NICKNAME_SIZE:copySize);
						strncpy(wip.fromMask,
								strchr(conn->buffer.str[0],'!')+1,
								WON_MAX_MASK_SIZE);
					}
				}

				for (x = 0; x < conn->buffer.words && x < 4; x++)
				{
					WON_HANDLERS *handlers = hashStringKeyGetValue(&ctx->ircHandlers,conn->buffer.str[x]);

					if (handlers)
					{
						WON_REGISTERED_HANDLER *reg = handlers->firstHandler;

						while (reg)
						{
							if (reg->position == x)
							{
								wip.modctx = reg->modctx;

								if ((retval = ((uint32 (*)(WON_IRC_PACKET *pkt))reg->function)(&wip)) == WON_STATUS_CLAIMPACKET)
									break;
								if (retval == WON_STATUS_CONNCLOSED)
									break;
							}

							reg = reg->next;
						}
					}
				}
			}
			break;
		case WON_CONNECTION_DCC:
			{
				WON_DCC_PACKET wdp;

				memset(&wdp,0,sizeof(WON_DCC_PACKET));

				wdp.ctx 	= ctx;
				wdp.buffer 	= &conn->buffer;
				wdp.dcc    	= wonDccGetConnectionByHandle(ctx,conn->fd);

				packet.pkt.dcc = &wdp;

				for (item = listGetFirstItem(&ctx->lowlevelRecvHandlers);
						item && item->value;
						item = listGetNextItem(item))
					if (((uint32 (*)(WON_MODULE_CTX *modctx,WON_PACKET *packet))item->id.longId)((WON_MODULE_CTX *)item->value,&packet) == WON_STATUS_FAILURE)
						break;

				if (item)
					break;
				
				for (x = 0; x < conn->buffer.words && x < 4; x++)
				{
					WON_HANDLERS *handlers = hashStringKeyGetValue(&ctx->dccHandlers,conn->buffer.str[x]);

					if (handlers)
					{
						WON_REGISTERED_HANDLER *reg = handlers->firstHandler;

						while (reg)
						{
							if (reg->position == x)
							{
								wdp.modctx = reg->modctx;

								if ((retval = ((uint32 (*)(WON_DCC_PACKET *pkt))reg->function)(&wdp)) == WON_STATUS_CLAIMPACKET)
									break;
								if (retval == WON_STATUS_CONNCLOSED)
									break;
							}

							reg = reg->next;
						}
					}
				}

			}
			break;
	}

	return retval;
}

void wonLoaderExit(WON_CTX *ctx,const char *fmt,...)
{
	va_list ap;
	char buf[512];

	va_start(ap,fmt);

	vsnprintf(buf,511,fmt,ap);
	
	va_end(ap);

	if (ctx->logfd)
		wonLog(ctx,NULL,"%s",buf);

	fprintf(stdout,"wonloader: exit -- %s\n",buf);

	exit(1);
}
